"""The strategy class."""

import datetime
import os
import pickle
import statistics
import uuid

import empyrical  # type: ignore
import numpy as np
import optuna
import pandas as pd
import pytz
from dateutil.relativedelta import relativedelta
from sklearn.metrics import precision_score  # type: ignore
from sklearn.metrics import accuracy_score, recall_score

from ..data.field_type import FieldType
from ..data.game_model import GAME_DT_COLUMN, GAME_WEEK_COLUMN
from ..data.league_model import DELIMITER
from .features import CombinedFeature
from .reducers import CombinedReducer
from .trainers import (FEATURES_USR_ATTR, HASH_USR_ATTR, CatboostTrainer,
                       VennAbersTrainer)
from .trainers.output_column import OUTPUT_COLUMN, OUTPUT_PROB_COLUMN_PREFIX

_SAMPLER_FILENAME = "sampler.pkl"


def _next_week_dt(
    current_dt: datetime.datetime | None, df: pd.DataFrame
) -> datetime.datetime | None:
    if df.empty:
        return None
    week_column = DELIMITER.join([GAME_WEEK_COLUMN])
    dt_column = DELIMITER.join([GAME_DT_COLUMN])
    if current_dt is not None:
        df = df[df[dt_column] >= current_dt]
    current_week = df.iloc[0][week_column]
    for _, row in df.iterrows():
        week = row[week_column]
        if current_week != week:
            return pd.to_datetime(row[dt_column]).to_pydatetime()
    try:
        return pd.to_datetime(df[dt_column]).to_pydatetime()[-1]  # type: ignore
    except AttributeError:
        return None


def _print_metrics(y_test: pd.DataFrame, y_pred: pd.DataFrame) -> float:
    accuracy = accuracy_score(y_test, y_pred)
    print(f"Accuracy: {accuracy}")
    print(f"Precision: {precision_score(y_test, y_pred)}")
    print(f"Recall: {recall_score(y_test, y_pred)}")
    return accuracy


class Strategy:
    """The strategy class."""

    # pylint: disable=too-many-locals

    _returns: pd.Series | None

    def __init__(self, df: pd.DataFrame, name: str) -> None:
        self._df = df
        self._name = name
        self._features = CombinedFeature()
        self._reducers = CombinedReducer(
            [
                DELIMITER.join([GAME_WEEK_COLUMN]),
                DELIMITER.join([GAME_DT_COLUMN]),
            ]
        )
        os.makedirs(name, exist_ok=True)
        storage_name = f"sqlite:///{name}/study.db"
        sampler_file = os.path.join(name, _SAMPLER_FILENAME)
        restored_sampler = None
        if os.path.exists(sampler_file):
            with open(sampler_file, "rb") as handle:
                restored_sampler = pickle.load(handle)
        self._study = optuna.create_study(
            study_name=name,
            storage=storage_name,
            load_if_exists=True,
            sampler=restored_sampler,
            direction=optuna.study.StudyDirection.MAXIMIZE,
        )
        self._returns = None

    def fit(self, start_dt: datetime.datetime | None = None):
        """Fits the strategy to the dataset by walking forward."""
        # pylint: disable=too-many-statements

        if start_dt is None:
            start_dt = max(
                self._df[[GAME_DT_COLUMN, x]]
                .dropna()[GAME_DT_COLUMN]
                .iloc[0]
                .to_pydatetime()
                for x in self._df.attrs[str(FieldType.ODDS)]
            )
            start_dt = max(
                start_dt,  # type: ignore
                pytz.utc.localize(datetime.datetime.now() - relativedelta(years=5)),
            )

        df = self._df.copy()
        df = df[
            df[GAME_DT_COLUMN]
            < pytz.utc.localize(datetime.datetime.now() - datetime.timedelta(days=1.0))
        ]
        training_cols = set(df.attrs[str(FieldType.POINTS)])
        x = self._features.process(df)
        x = self._reducers.process(x)
        y = df[list(training_cols)]
        y[OUTPUT_COLUMN] = np.argmax(y.to_numpy(), axis=1)
        if len(training_cols) == 2:
            y[OUTPUT_COLUMN] = y[OUTPUT_COLUMN].astype(bool)
        y = y[[OUTPUT_COLUMN]]

        # Walkforward by week
        predictions = []
        current_n_trials = max(32 - len(self._study.trials), 1)
        while True:
            start_dt = _next_week_dt(start_dt, x)
            if start_dt is None:
                break
            x_walk = x[x[GAME_DT_COLUMN] < start_dt]
            y_walk = y.iloc[: len(x_walk)]
            if len(x_walk) == len(x) or len(y_walk) == len(y):
                break
            folder = os.path.join(self._name, str(start_dt.date()))
            if os.path.exists(folder):
                continue
            os.makedirs(folder, exist_ok=True)
            print(f"Trainer {folder}")
            # next_dt = _next_week_dt(start_dt, x)
            x_test = x[x[GAME_DT_COLUMN] >= start_dt]
            y_test = y.iloc[len(x_walk) : len(x_walk) + len(x_test)]

            def objective(trial: optuna.Trial) -> float:
                trial.set_user_attr(HASH_USR_ATTR, str(uuid.uuid4()))
                trainer = VennAbersTrainer(
                    folder,
                    CatboostTrainer(
                        folder,
                        df.attrs[str(FieldType.CATEGORICAL)],
                        df.attrs[str(FieldType.TEXT)],
                        trial=trial,
                    ),
                )
                x_split, y_split = trainer.split_train_test(x_walk, y_walk)
                features, iterations = trainer.select_features(x_split, y_split)
                trial.set_user_attr(FEATURES_USR_ATTR, features)
                trial.set_user_attr("ITERATIONS", iterations)

                print("In Sample Metrics:")
                y_pred = trainer.predict(x_split[0])
                if y_pred is None:
                    raise ValueError("y_pred is null")
                _print_metrics(y_split[0], y_pred)

                print("Out of Sample Metrics:")
                y_pred = trainer.predict(x_test)
                if y_pred is None:
                    raise ValueError("y_pred is null")
                _print_metrics(y_test, y_pred)

                print("Test Metrics:")
                y_pred = trainer.predict(x_split[1])
                if y_pred is None:
                    raise ValueError("y_pred is null")
                return _print_metrics(y_split[1], y_pred)

            self._study.optimize(objective, n_trials=current_n_trials)
            current_n_trials = 1
            with open(os.path.join(self._name, _SAMPLER_FILENAME), "wb") as handle:
                pickle.dump(self._study.sampler, handle)
            best_trial = self._study.best_trial
            trainer = VennAbersTrainer(
                folder,
                CatboostTrainer(
                    folder,
                    df.attrs[str(FieldType.CATEGORICAL)],
                    df.attrs[str(FieldType.TEXT)],
                    trial=best_trial,
                ),
            )
            trainer.fit(
                (
                    x_walk[
                        [
                            x
                            for x in best_trial.user_attrs[FEATURES_USR_ATTR]
                            if x in x_walk.columns.values
                        ]
                    ],
                    None,
                ),
                (y_walk, None),
            )
            trainer.save()

            y_pred = trainer.predict(x_walk)
            if y_pred is None:
                raise ValueError("y_pred is null")
            print("Final In Sample Metrics:")
            _print_metrics(y_walk, y_pred)

            y_pred = trainer.predict(x_test)
            if y_pred is None:
                raise ValueError("y_pred is null")

            print("Final Out of Sample Metrics:")
            predictions.append(_print_metrics(y_test, y_pred))
        return statistics.mean(predictions)

    def predict(self, start_dt: datetime.datetime | None = None) -> pd.DataFrame:
        """Predict the results from walk-forward."""
        dt_column = DELIMITER.join([GAME_DT_COLUMN])

        if start_dt is None:
            start_dt = max(
                self._df[[dt_column, x]].dropna()[dt_column].iloc[0].to_pydatetime()
                for x in self._df.attrs[str(FieldType.ODDS)]
            )
            start_dt = max(
                start_dt,  # type: ignore
                pytz.utc.localize(datetime.datetime.now() - relativedelta(years=5)),
            )

        df = self._df.copy()
        df = df[
            df[dt_column]
            < pytz.utc.localize(datetime.datetime.now() - datetime.timedelta(days=1.0))
        ]
        training_cols = set(df.attrs[str(FieldType.POINTS)])
        x = self._features.process(df)
        y = df[list(training_cols)]
        y[OUTPUT_COLUMN] = np.argmax(y.to_numpy(), axis=1)
        if len(training_cols) == 2:
            y[OUTPUT_COLUMN] = y[OUTPUT_COLUMN].astype(bool)
        y = y[[OUTPUT_COLUMN]]

        while True:
            start_dt = _next_week_dt(start_dt, x)
            if start_dt is None:
                break
            x_walk = x[x[dt_column] < start_dt]
            if len(x_walk) == len(x):
                break
            next_dt = _next_week_dt(start_dt, x)
            if next_dt is None:
                break
            x_test = x[x[dt_column] >= start_dt]
            x_test = x_test[x_test[dt_column] < next_dt]
            folder = os.path.join(self._name, str(start_dt.date()))
            trainer = VennAbersTrainer(
                folder,
                CatboostTrainer(
                    folder,
                    self._df.attrs[str(FieldType.CATEGORICAL)],
                    self._df.attrs[str(FieldType.TEXT)],
                ),
            )
            trainer.load()
            y_prob = trainer.predict_proba(x_test)
            if y_prob is None:
                raise ValueError("y_prob is null")

            for column in y_prob.columns.values:
                if column not in x:
                    x[column] = None
                x_small = x[x[dt_column] >= start_dt]
                i_start = len(x) - len(x_small)
                x_small = x_small[x_small[dt_column] < next_dt]
                i_end = i_start + len(x_small)
                x.iloc[i_start:i_end, x.columns.get_loc(column)] = list(  # type: ignore
                    y_prob[column].values
                )

        df = df.reset_index().drop(columns=["index"])
        for points_col in df.attrs[str(FieldType.POINTS)]:
            x[points_col] = df[points_col].values

        return x

    def returns(self) -> pd.Series:
        """Render the returns of the strategy."""
        returns = self._returns
        if returns is None:
            df = self.predict()
            dt_column = DELIMITER.join([GAME_DT_COLUMN])
            points_cols = self._df.attrs[str(FieldType.POINTS)]

            def calculate_returns(kelly_ratio: float) -> pd.Series:
                index = []
                data = []
                for date, group in df.groupby([df[dt_column].dt.date]):
                    date = date[0]
                    index.append(date)

                    # Find the kelly criterion for each bet
                    fs = []
                    for _, row in group.iterrows():
                        row_df = row.to_frame().T
                        odds_df = row_df[self._df.attrs[str(FieldType.ODDS)]]
                        row_df = row_df[
                            [
                                x
                                for x in row_df.columns.values
                                if x.startswith(OUTPUT_PROB_COLUMN_PREFIX)
                            ]
                        ]
                        if row_df.isnull().values.any():
                            continue
                        arr = row_df.to_numpy().flatten()
                        team_idx = np.argmax(arr)
                        prob = arr[team_idx]
                        odds = list(
                            odds_df[
                                self._df.attrs[str(FieldType.ODDS)][team_idx]
                            ].values
                        )[0]
                        bet_prob = 1.0 / odds
                        f = max(prob - ((1.0 - prob) / bet_prob), 0.0) * kelly_ratio
                        fs.append(f)

                    # Make sure we aren't overallocating our capital
                    fs_sum = sum(fs)
                    if fs_sum > 1.0:
                        fs = [x / fs_sum for x in fs]

                    # Simulate the bets
                    bet_idx = 0
                    pl = 0.0
                    for _, row in group.iterrows():
                        row_df = row.to_frame().T
                        points_df = row_df[points_cols]
                        odds_df = row_df[self._df.attrs[str(FieldType.ODDS)]]
                        row_df = row_df[
                            [
                                x
                                for x in row_df.columns.values
                                if x.startswith(OUTPUT_PROB_COLUMN_PREFIX)
                            ]
                        ]
                        if row_df.isnull().values.any():
                            continue
                        arr = row_df.to_numpy().flatten()
                        team_idx = np.argmax(arr)
                        win_team_idx = np.argmax(points_df.to_numpy().flatten())
                        odds = list(
                            odds_df[
                                self._df.attrs[str(FieldType.ODDS)][team_idx]
                            ].values
                        )[0]
                        if team_idx == win_team_idx:
                            pl += odds * fs[bet_idx]
                        else:
                            pl -= fs[bet_idx]
                        bet_idx += 1

                    data.append(pl)

                return pd.Series(index=index, data=data, name=self._name)

            study = optuna.create_study(
                study_name="kelly",
                direction=optuna.study.StudyDirection.MAXIMIZE,
            )

            def objective(trial: optuna.Trial) -> float:
                ret = calculate_returns(trial.suggest_float("kelly_ratio", 0.0, 2.0))
                if abs(empyrical.max_drawdown(ret)) >= 1.0:
                    return 0.0
                return empyrical.calmar_ratio(ret)  # type: ignore

            study.optimize(objective, n_trials=100, show_progress_bar=True)

            returns = calculate_returns(
                study.best_trial.suggest_float("kelly_ratio", 0.0, 2.0)
            )
            self._returns = returns
        return returns



"""The total feature extractor."""

# pylint: disable=comparison-with-itself
import pandas as pd
import tqdm

from ...data.game_model import VENUE_COLUMN_PREFIX
from ...data.league_model import DELIMITER
from .columns import (find_player_count, find_team_count, player_column_prefix,
                      player_identifier_column, team_column_prefix,
                      team_identifier_column, venue_identifier_column)
from .feature import Feature

TOTAL_COLUMN_PREFIX = "total"
TOTAL_GAMES_COLUMN = "games"
TOTAL_KICKS_COLUMN = "kicks"


def _process_player_team_games(df: pd.DataFrame, cols: list[str]) -> pd.DataFrame:
    team_count = find_team_count(df)
    player_count = find_player_count(df, team_count)
    player_team_games: dict[str, int] = {}
    for i in range(team_count):
        for j in range(player_count):
            total_attendance_col = DELIMITER.join(
                [TOTAL_COLUMN_PREFIX, player_column_prefix(i, j), TOTAL_GAMES_COLUMN]
            )
            df[total_attendance_col] = None
    for row in tqdm.tqdm(df.itertuples(), desc="Processing player team games total"):
        for i in range(team_count):
            team_idx = row[cols.index(team_identifier_column(i)) + 1]
            # Check for NaNs
            if team_idx != team_idx:
                continue
            for j in range(player_count):
                player_col = player_identifier_column(i, j)
                if player_col not in cols:
                    continue
                player_idx = row[cols.index(player_col) + 1]
                # Check for NaNs
                if player_idx != player_idx:
                    continue
                key = "-".join([team_idx, player_idx])
                count = player_team_games.get(key, 0)
                total_attendance_col = DELIMITER.join(
                    [
                        TOTAL_COLUMN_PREFIX,
                        player_column_prefix(i, j),
                        TOTAL_GAMES_COLUMN,
                    ]
                )
                df.loc[row.Index, total_attendance_col] = count
                player_team_games[key] = count + 1
    return df


def _process_team_venue_games(df: pd.DataFrame, cols: list[str]) -> pd.DataFrame:
    venue_col = venue_identifier_column()
    if venue_col not in cols:
        return df
    team_count = find_team_count(df)
    for i in range(team_count):
        total_venue_col = DELIMITER.join(
            [
                TOTAL_COLUMN_PREFIX,
                team_column_prefix(i),
                VENUE_COLUMN_PREFIX,
                TOTAL_GAMES_COLUMN,
            ]
        )
        df[total_venue_col] = None
    venue_team_games: dict[str, int] = {}
    for row in tqdm.tqdm(df.itertuples(), desc="Processing team venue games total"):
        for i in range(team_count):
            team_idx = row[cols.index(team_identifier_column(i)) + 1]
            # Check for NaNs
            if team_idx != team_idx:
                continue
            venue_idx = row[cols.index(venue_identifier_column()) + 1]
            # Check for NaNs
            if venue_idx != venue_idx:
                continue
            key = "-".join([team_idx, venue_idx])
            count = venue_team_games.get(key, 0)
            total_venue_col = DELIMITER.join(
                [
                    TOTAL_COLUMN_PREFIX,
                    team_column_prefix(i),
                    VENUE_COLUMN_PREFIX,
                    TOTAL_GAMES_COLUMN,
                ]
            )
            df.loc[row.Index, total_venue_col] = count
            venue_team_games[key] = count + 1
    return df


class TotalFeature(Feature):
    """The total feature extractor class."""

    # pylint: disable=too-few-public-methods

    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """Process the dataframe and add the necessary features."""
        cols = df.columns.values.tolist()
        df = _process_player_team_games(df, cols)
        df = _process_team_venue_games(df, cols)
        return df.copy()



"""AFL AFLTables team model."""

import os
from urllib.parse import urlparse

import requests
from bs4 import BeautifulSoup

from ....cache import MEMORY
from ...team_model import TeamModel
from .afl_afltables_player_model import create_afl_afltables_player_model

_TEAM_NAME_MAP = {
    "Melbourne": ["ME"],
    "Geelong": ["GE"],
    "Fitzroy": ["FI"],
    "Collingwood": ["CW"],
    "Essendon": ["ES"],
    "South Melbourne": ["SM"],
    "St Kilda": ["SK"],
    "Carlton": ["CA"],
    "Sydney": ["SM", "SY"],
    "University": ["UN"],
    "Richmond": ["RI"],
    "North Melbourne": ["NM"],
    "Western Bulldogs": ["WB", "FO"],
    "Hawthorn": ["HW"],
    "Brisbane Bears": ["BB"],
    "West Coast": ["WC"],
    "Adelaide": ["AD"],
    "Fremantle": ["FR"],
    "Brisbane Lions": ["BL"],
    "Port Adelaide": ["PA"],
    "Gold Coast": ["GC"],
    "Greater Western Sydney": ["GW"],
}


@MEMORY.cache(ignore=["session"])
def create_afl_afltables_team_model(
    team_url: str,
    players: list[tuple[str, str, int | None]],
    points: float,
    session: requests.Session,
    last_ladder_ranks: dict[str, int] | None,
) -> TeamModel:
    """Create a team model from AFL Tables."""
    response = session.get(team_url)
    soup = BeautifulSoup(response.text, "html.parser")
    o = urlparse(team_url)
    last_component = o.path.split("/")[-1]
    identifier, _ = os.path.splitext(last_component)
    h1 = soup.find("h1")
    if h1 is None:
        raise ValueError("h1 is null.")
    name = h1.get_text()
    last_ladder_rank = None
    if last_ladder_ranks is not None and last_ladder_ranks:
        short_names = _TEAM_NAME_MAP[name]
        for short_name in short_names:
            if short_name in last_ladder_ranks:
                last_ladder_rank = last_ladder_ranks[short_name]
                break
    return TeamModel(
        identifier=identifier,
        name=name,
        players=[  # pyright: ignore
            create_afl_afltables_player_model(player_url, jersey, kicks)
            for player_url, jersey, kicks in players
        ],
        odds=[],
        points=points,
        ladder_rank=last_ladder_rank,
        location=None,
    )



"""The prototype class for odds."""

from pydantic import BaseModel, Field

from .bookie_model import BookieModel
from .field_type import TYPE_KEY, FieldType


class OddsModel(BaseModel):
    """The serialisable odds class."""

    odds: float = Field(..., json_schema_extra={TYPE_KEY: FieldType.ODDS})
    bookie: BookieModel



"""The prototype class for a bookie."""

from pydantic import BaseModel, Field

from .field_type import TYPE_KEY, FieldType


class BookieModel(BaseModel):
    """The serialisable bookie class."""

    identifier: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})
    name: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.TEXT})



"""The sportsball model trainers module."""

# ruff: noqa: F401
from .catboost import CatboostTrainer
from .trainer import FEATURES_USR_ATTR, HASH_USR_ATTR, Trainer
from .vennabers import VennAbersTrainer



"""AFL AFLTables player model."""

import os
from urllib.parse import urlparse

from ....cache import MEMORY
from ...player_model import PlayerModel


@MEMORY.cache
def create_afl_afltables_player_model(
    player_url: str, jersey: str, kicks: int | None
) -> PlayerModel:
    """Create a player model from AFL Tables."""
    o = urlparse(player_url)
    last_component = o.path.split("/")[-1]
    identifier, _ = os.path.splitext(last_component)
    jersey = "".join(filter(str.isdigit, jersey))
    return PlayerModel(identifier=identifier, jersey=jersey, kicks=kicks)



"""ESPN team model."""

from typing import Any

from ...cache import MEMORY
from ..odds_model import OddsModel
from ..team_model import TeamModel
from .espn_player_model import create_espn_player_model


@MEMORY.cache
def create_espn_team_model(
    team: dict[str, Any],
    roster_dict: dict[str, Any],
    odds: list[OddsModel],
    score_dict: dict[str, Any],
) -> TeamModel:
    """Create team model from ESPN."""
    identifier = team["id"]
    name = team.get("name", team.get("fullName", team.get("displayName")))
    if name is None:
        raise ValueError("name is null")
    location = team["location"]
    players = []
    for entity in roster_dict.get("entries", []):
        player = create_espn_player_model(entity)
        players.append(player)
    points = score_dict["value"]
    return TeamModel(
        identifier=identifier,
        name=name,
        location=location,
        players=players,
        odds=odds,
        points=points,
        ladder_rank=None,
    )



"""NBA API player model."""

import json

import pandas as pd
from nba_api.stats.endpoints import commonplayerinfo  # type: ignore

from ....cache import MEMORY
from ...player_model import PlayerModel


@MEMORY.cache
def create_nba_nba_player_model(
    row: pd.Series, player_index: pd.DataFrame
) -> PlayerModel:
    """Create a player model from NBA API."""
    jersey = None
    player_id = str(row["PERSON_ID"])
    player_index_df = player_index[player_index["PERSON_ID"].astype(str) == player_id]
    if not player_index_df.empty:
        jersey = player_index_df.iloc[0]["JERSEY_NUMBER"]
    else:
        try:
            cpi = commonplayerinfo.CommonPlayerInfo(player_id=player_id)
            cpi_row = cpi.common_player_info.get_data_frame().squeeze(axis=0)
            jersey = cpi_row["JERSEY"]
        except json.decoder.JSONDecodeError:
            pass
    return PlayerModel(identifier=player_id, jersey=jersey, kicks=None)



"""The prototype class for a trainer."""

import hashlib
import os
from typing import Any

import optuna
import pandas as pd
from sklearn.model_selection import train_test_split  # type: ignore

HASH_USR_ATTR = "HASH"
FEATURES_USR_ATTR = "FEATURES"


def _hash_df(df: pd.DataFrame) -> int:
    return int(
        hashlib.sha256(
            pd.util.hash_pandas_object(df, index=True).values,  # type: ignore
        ).hexdigest(),
        16,
    )


class Trainer:
    """The prototype trainer class."""

    def __init__(
        self,
        folder: str,
        trial: optuna.trial.Trial | optuna.trial.FrozenTrial | None = None,
    ) -> None:
        self._folder = folder
        self._test_size = 0.2
        self._trial = trial
        if trial is not None:
            self._test_size = trial.suggest_float("test_size", 0.0, 0.5)

    @property
    def clf(self) -> Any:
        """The underlying classifier"""
        raise NotImplementedError("clf is not implemented in parent class.")

    @property
    def salt(self) -> str:
        """The salt to use when hashing the predictions."""
        raise NotImplementedError("salt is not implemented in parent class.")

    def fit(
        self,
        x: tuple[pd.DataFrame, pd.DataFrame | None],
        y: tuple[pd.DataFrame, pd.DataFrame | None],
    ):
        """Fit the data."""
        raise NotImplementedError("fit is not implemented in parent class.")

    def save(self):
        """Save the trainer."""
        raise NotImplementedError("save is not implemented in parent class.")

    def load(self):
        """Load the trainer."""
        raise NotImplementedError("load is not implemented in parent class.")

    def predict(self, x: pd.DataFrame) -> pd.DataFrame | None:
        """Predict the Y values."""
        filename = os.path.join(self._folder, f"{self.salt}_{_hash_df(x)}.parquet")
        if os.path.exists(filename):
            return pd.read_parquet(filename)
        return None

    def predict_proba(self, x: pd.DataFrame) -> pd.DataFrame | None:
        """Predict the Y probabilities."""
        filename = os.path.join(
            self._folder, f"{self.salt}_{_hash_df(x)}_proba.parquet"
        )
        if os.path.exists(filename):
            return pd.read_parquet(filename)
        return None

    def save_prediction(self, x: pd.DataFrame, y: pd.DataFrame):
        """Save the prediction for later."""
        filename = os.path.join(self._folder, f"{self.salt}_{_hash_df(x)}.parquet")
        y.to_parquet(filename)

    def save_prediction_proba(self, x: pd.DataFrame, y: pd.DataFrame):
        """Save the prediction probabilities for later."""
        filename = os.path.join(
            self._folder, f"{self.salt}_{_hash_df(x)}_proba.parquet"
        )
        y.to_parquet(filename)

    def select_features(
        self,
        x: tuple[pd.DataFrame, pd.DataFrame | None],
        y: tuple[pd.DataFrame, pd.DataFrame | None],
    ) -> tuple[list[str], int]:
        """Select the features from the training data."""
        raise NotImplementedError("select_features is not implemented in parent class.")

    def split_train_test(
        self, x: pd.DataFrame, y: pd.DataFrame
    ) -> tuple[tuple[pd.DataFrame, pd.DataFrame], tuple[pd.DataFrame, pd.DataFrame]]:
        """Splits the x/y data into train and test sets."""
        x_train, x_test, y_train, y_test = train_test_split(
            x, y, test_size=self._test_size, shuffle=False
        )
        return (x_train, x_test), (y_train, y_test)  # type: ignore



"""A weight based on the prediction class numbers."""

import numpy as np
import pandas as pd
from sklearn.utils.class_weight import compute_class_weight  # type: ignore

from .weight import Weight


class ClassWeight(Weight):
    """Class weight class."""

    @classmethod
    def name(cls) -> str:
        """The name of the weight class."""
        return "class"

    def process(self, y: pd.DataFrame) -> np.ndarray:
        arr = y.astype(int).to_numpy().flatten().astype(float)
        unique_vals = np.unique(arr)
        w_arr = compute_class_weight(
            class_weight="balanced", classes=unique_vals, y=arr
        )
        for count, unique_val in enumerate(unique_vals):
            arr[arr == unique_val] = w_arr[count]
        return arr



"""The skill feature extractor."""

import datetime
from warnings import simplefilter

import pandas as pd
import tqdm
from dateutil.relativedelta import relativedelta
from openskill.models import PlackettLuce, PlackettLuceRating
from pandarallel import pandarallel  # type: ignore

from ...data.field_type import FieldType
from ...data.game_model import GAME_DT_COLUMN
from ...data.league_model import DELIMITER
from .columns import (find_player_count, find_team_count, player_column_prefix,
                      player_identifier_column, team_column_prefix,
                      team_identifier_column, team_points_column)
from .feature import Feature

SKILL_COLUMN_PREFIX = "skill"
SKILL_MU_COLUMN = "mu"
SKILL_SIGMA_COLUMN = "sigma"
SKILL_RANKING_COLUMN = "ranking"
SKILL_PROBABILITY_COLUMN = "probability"
YEAR_SLICE_ALL = "all"


def _slice_df(
    df: pd.DataFrame, date: datetime.date, year_slice: int | None
) -> pd.DataFrame:
    df_slice = df[df[GAME_DT_COLUMN].dt.date < date]
    if year_slice is not None:
        start_date = date - relativedelta(years=year_slice)
        df_slice = df_slice[df_slice[GAME_DT_COLUMN].dt.date > start_date]
    return df_slice


def _create_teams(
    df_slice: pd.DataFrame, team_count: int, group: pd.DataFrame
) -> tuple[PlackettLuce, dict[str, PlackettLuceRating]]:
    team_model = PlackettLuce()
    teams = {}
    for i in range(team_count):
        for team_identifier in df_slice[team_identifier_column(i)].unique():
            if not isinstance(team_identifier, str):
                continue
            if team_identifier in teams:
                continue
            teams[team_identifier] = team_model.rating(name=team_identifier)
        for team_identifier in group[team_identifier_column(i)].unique():
            if not isinstance(team_identifier, str):
                continue
            if team_identifier in teams:
                continue
            teams[team_identifier] = team_model.rating(name=team_identifier)
    return team_model, teams


def _create_player_teams(
    df_slice: pd.DataFrame, team_count: int, player_count: int, group: pd.DataFrame
) -> tuple[PlackettLuce, dict[str, PlackettLuceRating]]:
    player_model = PlackettLuce()
    players = {}
    for i in range(player_count):
        for j in range(team_count):
            if player_identifier_column(j, i) in df_slice.columns.values:
                for player_identifier in df_slice[
                    player_identifier_column(j, i)
                ].unique():
                    if not isinstance(player_identifier, str):
                        continue
                    if player_identifier in players:
                        continue
                    players[player_identifier] = player_model.rating(
                        name=player_identifier
                    )
            if player_identifier_column(j, i) in group.columns.values:
                for player_identifier in group[player_identifier_column(j, i)].unique():
                    if not isinstance(player_identifier, str):
                        continue
                    if player_identifier in players:
                        continue
                    players[player_identifier] = player_model.rating(
                        name=player_identifier
                    )
    return player_model, players


def _find_matches(
    row: pd.Series,
    team_count: int,
    player_count: int,
    teams: dict[str, PlackettLuceRating],
    players: dict[str, PlackettLuceRating],
) -> tuple[list[float], list[list[PlackettLuceRating]], list[list[PlackettLuceRating]]]:
    points = []
    team_match = []
    player_match = []
    for i in range(team_count):
        points_col = team_points_column(i)
        if pd.isnull(row[points_col]):
            continue
        points.append(float(row[points_col]))
        team_idx = row[team_identifier_column(i)]
        team_match.append([teams[team_idx]])
        player_team = []
        for j in range(player_count):
            try:
                if pd.isnull(row[player_identifier_column(i, j)]):
                    continue
            except KeyError:
                continue
            player_team.append(players[row[player_identifier_column(i, j)]])
        player_match.append(player_team)
    return points, team_match, player_match


def _rate_match(
    model: PlackettLuce,
    match: list[list[PlackettLuceRating]],
    points: list[float],
    ratings: dict[str, PlackettLuceRating],
) -> dict[str, PlackettLuceRating]:
    if not match:
        return ratings
    for team in match:
        if not team:
            return ratings
    output = model.rate(match, scores=points)
    for team in output:
        for player in team:
            name = player.name
            if name is None:
                continue
            ratings[name] = player
    return ratings


def _simulate_games(
    df_slice: pd.DataFrame,
    team_count: int,
    player_count: int,
    model_team: tuple[PlackettLuce, dict[str, PlackettLuceRating]],
    model_players: tuple[PlackettLuce, dict[str, PlackettLuceRating]],
) -> tuple[
    tuple[PlackettLuce, dict[str, PlackettLuceRating]],
    tuple[PlackettLuce, dict[str, PlackettLuceRating]],
]:
    team_model, teams = model_team
    player_model, players = model_players

    def _simulate_match(row: pd.Series) -> pd.Series:
        nonlocal team_count
        nonlocal player_count
        nonlocal teams
        nonlocal players
        nonlocal team_model
        nonlocal player_model
        points, team_match, player_match = _find_matches(
            row, team_count, player_count, teams, players
        )
        teams = _rate_match(team_model, team_match, points, teams)
        players = _rate_match(player_model, player_match, points, players)
        return row

    df_slice.apply(_simulate_match, axis=1)

    return (team_model, teams), (player_model, players)


def _find_player_team(
    player_count: int,
    players: dict[str, PlackettLuceRating],
    row: pd.Series,
    team_index: int,
) -> list[PlackettLuceRating]:
    player_team = []
    for j in range(player_count):
        try:
            if pd.isnull(row[player_identifier_column(team_index, j)]):
                continue
        except KeyError:
            continue
        player_idx = row[player_identifier_column(team_index, j)]
        player = players[player_idx]
        player_team.append(player)
    return player_team


def _find_team_team(
    team_index: int, row: pd.Series, year_col: str, teams: dict[str, PlackettLuceRating]
) -> tuple[list[PlackettLuceRating], pd.Series]:
    team_col_prefix = team_column_prefix(team_index)
    team_idx_col = team_identifier_column(team_index)
    if pd.isnull(row[team_idx_col]):
        return [], row
    team_idx = row[team_idx_col]
    team_skill_col_prefix = DELIMITER.join(
        [team_col_prefix, SKILL_COLUMN_PREFIX, year_col]
    )
    team_mu_col = DELIMITER.join([team_skill_col_prefix, SKILL_MU_COLUMN])
    team_sigma_col = DELIMITER.join([team_skill_col_prefix, SKILL_SIGMA_COLUMN])
    team = teams[team_idx]
    row[team_mu_col] = team.mu
    row[team_sigma_col] = team.sigma
    return [team], row


def _find_row_matches(
    row: pd.Series,
    counts: tuple[int, int],
    year_col: str,
    teams: dict[str, PlackettLuceRating],
    players: dict[str, PlackettLuceRating],
) -> tuple[pd.Series, list[list[PlackettLuceRating]], list[list[PlackettLuceRating]]]:
    team_count, player_count = counts
    team_match = []
    player_match = []
    for i in range(team_count):
        team, row = _find_team_team(i, row, year_col, teams)
        if not team:
            continue
        team_match.append(team)
        player_team = _find_player_team(player_count, players, row, i)
        player_match.append(player_team)
    return row, team_match, player_match


def _rank_team_predictions(
    team_match: list[list[PlackettLuceRating]],
    row: pd.Series,
    team_model: PlackettLuce,
    year_col: str,
) -> pd.Series:
    rank_team_predictions = team_model.predict_rank(team_match)
    for i, (rank, prob) in enumerate(rank_team_predictions):
        team_skill_col_prefix = DELIMITER.join(
            [
                team_column_prefix(i),
                SKILL_COLUMN_PREFIX,
                year_col,
            ]
        )
        team_ranking_col = DELIMITER.join([team_skill_col_prefix, SKILL_RANKING_COLUMN])
        team_prob_col = DELIMITER.join(
            [team_skill_col_prefix, SKILL_PROBABILITY_COLUMN]
        )
        row[team_ranking_col] = rank
        row[team_prob_col] = prob
    return row


def _rank_player_predictions(
    row: pd.Series,
    player_model: PlackettLuce,
    player_match: list[list[PlackettLuceRating]],
    year_col: str,
) -> pd.Series:
    for match in player_match:
        if not match:
            return row
    rank_player_predictions = player_model.predict_rank(player_match)
    for i, (rank, prob) in enumerate(rank_player_predictions):
        player_skill_col_prefix = DELIMITER.join(
            [
                player_column_prefix(i, None),
                SKILL_COLUMN_PREFIX,
                year_col,
            ]
        )
        player_ranking_col = DELIMITER.join(
            [player_skill_col_prefix, SKILL_RANKING_COLUMN]
        )
        player_prob_col = DELIMITER.join(
            [player_skill_col_prefix, SKILL_PROBABILITY_COLUMN]
        )
        row[player_ranking_col] = rank
        row[player_prob_col] = prob
    return row


def _create_feature_cols(
    group: pd.DataFrame,
    year_slice: int | None,
    counts: tuple[int, int],
    model_team: tuple[PlackettLuce, dict[str, PlackettLuceRating]],
    model_players: tuple[PlackettLuce, dict[str, PlackettLuceRating]],
) -> pd.DataFrame:
    year_col = str(year_slice)
    team_model, teams = model_team
    player_model, players = model_players
    team_count, player_count = counts

    def _apply_group_skills_features(row: pd.Series) -> pd.Series:
        row, team_match, player_match = _find_row_matches(
            row, (team_count, player_count), year_col, teams, players
        )
        row = _rank_team_predictions(team_match, row, team_model, year_col)
        row = _rank_player_predictions(row, player_model, player_match, year_col)
        return row

    return group.apply(_apply_group_skills_features, axis=1)


def _create_all_features(
    df: pd.DataFrame, team_count: int, player_count: int
) -> pd.DataFrame:
    team_model, teams = _create_teams(df, team_count, df)
    player_model, players = _create_player_teams(df, team_count, player_count, df)

    def _apply_skills_features(row: pd.Series) -> pd.Series:
        nonlocal team_model
        nonlocal teams
        nonlocal player_model
        nonlocal players
        row, team_match, player_match = _find_row_matches(
            row, (team_count, player_count), YEAR_SLICE_ALL, teams, players
        )
        (team_model, teams), (player_model, players) = _simulate_games(
            row.to_frame().T,
            team_count,
            player_count,
            (team_model, teams),
            (player_model, players),
        )
        row = _rank_team_predictions(team_match, row, team_model, YEAR_SLICE_ALL)
        row = _rank_player_predictions(row, player_model, player_match, YEAR_SLICE_ALL)
        return row

    df = df.apply(_apply_skills_features, axis=1)
    return df


class SkillFeature(Feature):
    """The skill feature extractor class."""

    # pylint: disable=too-few-public-methods

    def __init__(self, year_slices: list[int | None]) -> None:
        super().__init__()
        self._year_slices = year_slices
        tqdm.tqdm.pandas()
        simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
        pandarallel.initialize(progress_bar=True)

    def _create_columns(self, df: pd.DataFrame, team_count: int) -> pd.DataFrame:
        golden_features = df.attrs.get(str(FieldType.GOLDEN), [])
        for year_slice in self._year_slices:
            year_col = str(year_slice) if year_slice is not None else YEAR_SLICE_ALL
            for i in range(team_count):
                for team_col_suffix in [
                    SKILL_MU_COLUMN,
                    SKILL_SIGMA_COLUMN,
                    SKILL_RANKING_COLUMN,
                    SKILL_PROBABILITY_COLUMN,
                ]:
                    team_col = DELIMITER.join(
                        [
                            team_column_prefix(i),
                            SKILL_COLUMN_PREFIX,
                            year_col,
                            team_col_suffix,
                        ]
                    )
                    df[team_col] = 0.0
                    golden_features.append(team_col)
                for empty_player_col_suffix in [
                    SKILL_RANKING_COLUMN,
                    SKILL_PROBABILITY_COLUMN,
                ]:
                    player_col = DELIMITER.join(
                        [
                            player_column_prefix(i, None),
                            SKILL_COLUMN_PREFIX,
                            year_col,
                            empty_player_col_suffix,
                        ]
                    )
                    df[player_col] = 0.0
                    golden_features.append(player_col)
        df.attrs[str(FieldType.GOLDEN)] = golden_features
        return df.copy()

    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """Process the dataframe and add the necessary features."""
        print(f"Calculating Skills Features over the year slices: {self._year_slices}")
        team_count = find_team_count(df)
        player_count = find_player_count(df, team_count)
        df = self._create_columns(df, team_count)
        if None in self._year_slices:
            df = _create_all_features(df, team_count, player_count)

        def calculate_skills(group: pd.DataFrame) -> pd.DataFrame:
            nonlocal df
            nonlocal team_count
            nonlocal player_count
            dates = group[GAME_DT_COLUMN].dt.date.values.tolist()
            if not dates:
                return group
            date = dates[0]
            for year_slice in self._year_slices:
                if year_slice is None:
                    continue
                df_slice = _slice_df(df, date, year_slice)  # type: ignore
                if df_slice.empty:
                    continue
                team_model, teams = _create_teams(df_slice, team_count, group)
                player_model, players = _create_player_teams(
                    df_slice, team_count, player_count, group
                )
                (team_model, teams), (player_model, players) = _simulate_games(
                    df_slice,
                    team_count,
                    player_count,
                    (team_model, teams),
                    (player_model, players),
                )
                group = _create_feature_cols(
                    group,
                    year_slice,
                    (team_count, player_count),
                    (team_model, teams),
                    (player_model, players),
                )
            return group

        return (
            df.groupby(  # type: ignore
                [df[GAME_DT_COLUMN].dt.date]
            )
            .parallel_apply(calculate_skills)
            .reset_index(drop=True)
        )



"""A weight based on exponential bias towards the current time."""

import numpy as np
import pandas as pd

from .weight import Weight


class ExponentialWeight(Weight):
    """Exponential weight class."""

    @classmethod
    def name(cls) -> str:
        """The name of the weight class."""
        return "exponential"

    def process(self, y: pd.DataFrame) -> np.ndarray:
        return np.linspace(0, 1.0, len(y)) ** 2.0



"""Output column."""

OUTPUT_COLUMN = "output"
OUTPUT_PROB_COLUMN_PREFIX = "output_prob_"


def output_prob_column(class_idx: int) -> str:
    """Produce a column name for the output probability for a class."""
    return f"{OUTPUT_PROB_COLUMN_PREFIX}{class_idx}"



"""The main module for sportsball."""
# ruff: noqa: F401

__VERSION__ = "0.1.1"

from .data import (AddressModel, BookieModel, GameModel, League, LeagueModel,
                   PlayerModel, SeasonType, TeamModel, VenueModel)
from .portfolio import Portfolio
from .sportsball import SportsBall
from .strategy import Strategy



"""The prototype class for a player."""

from typing import Literal

from pydantic import BaseModel, Field

from .field_type import TYPE_KEY, FieldType

PLAYER_KICKS_COLUMN: Literal["kicks"] = "kicks"
PLAYER_IDENTIFIER_COLUMN: Literal["identifier"] = "identifier"


class PlayerModel(BaseModel):
    """The serialisable player class."""

    identifier: str = Field(
        ...,
        json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL},
        alias=PLAYER_IDENTIFIER_COLUMN,
    )
    jersey: str | None = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})
    kicks: int | None = Field(
        ...,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=PLAYER_KICKS_COLUMN,
    )



"""The enumeration of the different supported leagues."""

from enum import StrEnum


class League(StrEnum):
    """An enumeration over the different leagues."""

    AFL = "afl"
    NBA = "nba"
    NCAAF = "ncaaf"
    NFL = "nfl"



"""The portfolio class."""

import datetime
import logging
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyfolio as pf  # type: ignore
from fullmonte import plot, simulate  # type: ignore
from skfolio import RiskMeasure  # type: ignore
from skfolio.optimization import MeanRisk, ObjectiveFunction  # type: ignore

from ..strategy import Strategy


class Portfolio:
    """The portfolio class."""

    def __init__(self, strategies: list[Strategy], name: str) -> None:
        self._name = name
        self._strategies = strategies
        os.makedirs(name, exist_ok=True)

    def fit(self) -> pd.DataFrame:
        """Fits the portfolio to the strategies."""
        # pylint: disable=unsubscriptable-object
        returns = pd.DataFrame([x.returns() for x in self._strategies]).T.fillna(0.0)
        returns.index = pd.to_datetime(returns.index)

        # Walkforward sharpe optimization
        ret = returns.copy()
        ret[self._name] = np.NaN
        for index in returns.index:
            dt = index
            x = returns[returns.index < dt]
            if x.empty or len(np.unique(x)) < 10:
                ret.loc[index, self._name] = (
                    returns.loc[index] * (1.0 / len(returns.columns.values))
                ).sum()
            else:
                model = MeanRisk(
                    risk_measure=RiskMeasure.VARIANCE,
                    objective_function=ObjectiveFunction.MAXIMIZE_RETURN,
                    portfolio_params={"name": "Max Sharpe"},
                )
                try:
                    model.fit(x.to_numpy())
                    ret.loc[index, self._name] = (
                        returns.loc[index] * model.weights_
                    ).sum()
                except ValueError as e:
                    logging.warning(
                        "Encountered %s when fitting meanrisk model.", str(e)
                    )
                    ret.loc[index, self._name] = (
                        returns.loc[index] * (1.0 / len(returns.columns.values))
                    ).sum()

        ret = ret.asfreq("D").fillna(0.0)
        ret.index = ret.index.tz_localize("UTC")  # type: ignore
        return ret

    def render(
        self,
        returns: pd.DataFrame,
        start_money: float = 100000.0,
        from_date: datetime.datetime | None = None,
    ):
        """Renders the statistics of the portfolio."""

        def render_series(series: pd.Series) -> None:
            pf.create_full_tear_sheet(series)
            plt.savefig(os.path.join(self._name, f"{col}_tear_sheet.png"), dpi=300)
            ret = np.concatenate(
                (np.array([start_money]), series.to_numpy().flatten() + 1.0)
            ).cumprod()
            plot(simulate(pd.Series(ret)))
            plt.savefig(os.path.join(self._name, f"{col}_monte_carlo.png"), dpi=300)
            log_series = pd.Series(data=np.log(ret), index=series.index)
            log_series.plot()
            plt.savefig(os.path.join(self._name, f"{col}_log_returns.png"), dpi=300)

        if from_date is not None:
            returns = returns.loc[returns.index.date >= from_date]  # type: ignore
        for col in returns.columns.values:
            series = returns[col]
            series = series[
                series.index >= series.where(series != 0.0).first_valid_index()
            ]
            render_series(series)



"""NFL SportsDB team model."""

from ....cache import MEMORY
from ...team_model import TeamModel


@MEMORY.cache
def create_nfl_sportsdb_team_model(team_id: str, name: str, points: float) -> TeamModel:
    """Create a team model based off the sportsdb NFL response."""
    return TeamModel(
        identifier=team_id,
        name=name,
        points=points,
        players=[],
        odds=[],
        ladder_rank=None,
        location=None,
    )



"""NBA combined league model."""

import requests

from ...combined.combined_league_model import CombinedLeagueModel
from ...league import League
from ..espn.nba_espn_league_model import NBAESPNLeagueModel
from ..nba.nba_nba_league_model import NBANBALeagueModel

ORLANDO_MAGIC = "1610612753"
ATLANTA_HAWKS = "1610612737"
PORTLAND_TRAIL_BLAZERS = "1610612757"
DENVER_NUGGETS = "1610612743"
MIAMI_HEAT = "1610612748"
LOS_ANGELES_LAKERS = "1610612747"
BOSTON_CELTICS = "1610612738"
OKLAHOMA_CITY_THUNDER = "1610612760"
SACRAMENTO_KINGS = "1610612758"
INDIANA_PACERS = "1610612754"
SAN_ANTONIO_SPURS = "1610612759"
PHOENIX_SUNS = "1610612756"
WASHINGTON_WIZARDS = "1610612764"
PHILADELPHIA_76ERS = "1610612755"
TORONTO_RAPTORS = "1610612761"
GOLDEN_STATE_WARRIORS = "1610612744"
MINNESOTA_TIMBERWOLVES = "1610612750"
HOUSTON_ROCKETS = "1610612745"
MEMPHIS_GRIZZLIES = "1610612763"
NEW_YORK_KNICKS = "1610612752"
CHARLOTTE_HORNETS = "1610612766"
BROOKLYN_NETS = "1610612751"
DALLAS_MAVERICKS = "1610612742"
CLEVELAND_CAVALIERS = "1610612739"
MILWAUKEE_BUCKS = "1610612749"
DETROIT_PISTONS = "1610612765"
NEW_ORLEANS_PELICANS = "1610612740"
CHICAGO_BULLS = "1610612741"
LA_CLIPPERS = "1610612746"
UTAH_JAZZ = "1610612762"
UNKNOWN = "1610616842"
TEAM_JALEN = "1610616849"
TEAM_TAMIKA = "1610616850"
TEAM_PAU = "1610616847"
TEAM_DETLEF = "1610616848"
TEAM_STEPHEN_A = "1610616840"
EAST_NBA_ALL_STARS = "1610616833"
WEST_NBA_ALL_STARS = "1610616834"
TEAM_SHANNON = "1610616839"
NEW_ZEALAND_BREAKERS = "15020"
RATIOPHARM_ULM = "15025"
CAIRNS_TAIPANS = "15022"
RAANANA_MACCABI = "50009"
FLAMENGO = "12325"
MADRID_BALONCESTO = "12315"
ADELAIDE_36ERS = "15019"
TEAM_USA = "1610616843"
SHANGHAI_SHARKS = "12329"
SESI_FRANCA_BASKETBALL_CLUB_OF_BRAZIL = "12332"
HAIFA_MACABI = "93"
BUENOS_AIRES_SAN_LORENZO = "12330"
GUANGZHOU_LONG_LIONS = "15018"
MELBOURNE_UNITED = "15016"
WORLD = "1610616844"
BEIJING_DUCKS = "15021"
PERTH_WILDCATS = "104"
SYDNEY_KINGS = "15015"
BRISBANE_BULLETS = "15017"
BARCELONA_LASSA = "12304"
ISTANBUL_FENERBAHCE_SPORTS_CLUB = "12321"
MILANO_OLIMPIA = "94"
BAURU_PASCHOALOTTO = "12328"
TEL_AVIV_MACCABI_ELECTRA = "41"
BERLIN_ALBA = "12323"
BILBAO_BASKET = "12324"
MOSCOW_CSKA = "12308"
SIENA_MONTEPASCI = "12322"
SOPHOMORES = "1610616838"
ROOKIES = "1610616837"
VITORIA_CAJAL_LABORAL = "95"
BELGRADE_PARTIZAN = "12319"
ATHENS_OLYMPIACOS = "12320"
HORNETS = "1610612778"
VILNIUS_LIETUVOS_RYTAS = "12314"
ISTANBUL_EFES_PILSEN = "12309"
LOTTOMATICA_ROMA = "12317"
UNICAJA_MALAGA = "12311"
MMT_ESTUDIANTES = "12318"
PANATHINAIKOS = "12313"
ZALGIRIS_KAUNAS = "12316"
TEAM_CHINA = "45"
CLUTCH_CITY = "1610612767"
H_TOWN = "1610612778"
LYON_VILLEURBANNE_ADECCO_ASVEL = "12305"
KHIMKI_BC = "12306"
TEL_AVIV_MACCABI_ELITE = "12303"
NBA_TEAM_IDENTITY_MAP = {
    "1610612753": ORLANDO_MAGIC,
    "1610612737": ATLANTA_HAWKS,
    "1610612757": PORTLAND_TRAIL_BLAZERS,
    "1610612743": DENVER_NUGGETS,
    "1610612748": MIAMI_HEAT,
    "1610612747": LOS_ANGELES_LAKERS,
    "1610612738": BOSTON_CELTICS,
    "1610612760": OKLAHOMA_CITY_THUNDER,
    "1610612758": SACRAMENTO_KINGS,
    "1610612754": INDIANA_PACERS,
    "1610612759": SAN_ANTONIO_SPURS,
    "1610612756": PHOENIX_SUNS,
    "1610612764": WASHINGTON_WIZARDS,
    "1610612755": PHILADELPHIA_76ERS,
    "1610612761": TORONTO_RAPTORS,
    "1610612744": GOLDEN_STATE_WARRIORS,
    "1610612750": MINNESOTA_TIMBERWOLVES,
    "1610612745": HOUSTON_ROCKETS,
    "1610612763": MEMPHIS_GRIZZLIES,
    "1610612752": NEW_YORK_KNICKS,
    "1610612766": CHARLOTTE_HORNETS,
    "1610612751": BROOKLYN_NETS,
    "1610612742": DALLAS_MAVERICKS,
    "1610612739": CLEVELAND_CAVALIERS,
    "1610612749": MILWAUKEE_BUCKS,
    "1610612765": DETROIT_PISTONS,
    "1610612740": NEW_ORLEANS_PELICANS,
    "1610612741": CHICAGO_BULLS,
    "1610612746": LA_CLIPPERS,
    "1610612762": UTAH_JAZZ,
    "1610616842": UNKNOWN,
    "1610616841": UNKNOWN,
    "1610616849": TEAM_JALEN,
    "1610616850": TEAM_TAMIKA,
    "1610616847": TEAM_PAU,
    "1610616848": TEAM_DETLEF,
    "1610616840": TEAM_STEPHEN_A,
    "1610616833": EAST_NBA_ALL_STARS,
    "1610616834": WEST_NBA_ALL_STARS,
    "1610616839": TEAM_SHANNON,
    "15020": NEW_ZEALAND_BREAKERS,
    "15025": RATIOPHARM_ULM,
    "15022": CAIRNS_TAIPANS,
    "50009": RAANANA_MACCABI,
    "12325": FLAMENGO,
    "12315": MADRID_BALONCESTO,
    "15019": ADELAIDE_36ERS,
    "1610616843": TEAM_USA,
    "12329": SHANGHAI_SHARKS,
    "12332": SESI_FRANCA_BASKETBALL_CLUB_OF_BRAZIL,
    "93": HAIFA_MACABI,
    "12330": BUENOS_AIRES_SAN_LORENZO,
    "15018": GUANGZHOU_LONG_LIONS,
    "15016": MELBOURNE_UNITED,
    "1610616844": WORLD,
    "15021": BEIJING_DUCKS,
    "104": PERTH_WILDCATS,
    "15015": SYDNEY_KINGS,
    "15017": BRISBANE_BULLETS,
    "12304": BARCELONA_LASSA,
    "12321": ISTANBUL_FENERBAHCE_SPORTS_CLUB,
    "94": MILANO_OLIMPIA,
    "12328": BAURU_PASCHOALOTTO,
    "41": TEL_AVIV_MACCABI_ELECTRA,
    "12323": BERLIN_ALBA,
    "12324": BILBAO_BASKET,
    "12308": MOSCOW_CSKA,
    "12322": SIENA_MONTEPASCI,
    "1610616838": SOPHOMORES,
    "1610616837": ROOKIES,
    "95": VITORIA_CAJAL_LABORAL,
    "12319": BELGRADE_PARTIZAN,
    "12320": ATHENS_OLYMPIACOS,
    "1610612778": HORNETS,
    "12314": VILNIUS_LIETUVOS_RYTAS,
    "12309": ISTANBUL_EFES_PILSEN,
    "12317": LOTTOMATICA_ROMA,
    "12311": UNICAJA_MALAGA,
    "12318": MMT_ESTUDIANTES,
    "12313": PANATHINAIKOS,
    "12316": ZALGIRIS_KAUNAS,
    "45": TEAM_CHINA,
    "1610612767": CLUTCH_CITY,
    "12305": LYON_VILLEURBANNE_ADECCO_ASVEL,
    "12306": KHIMKI_BC,
    "12303": TEL_AVIV_MACCABI_ELITE,
    "12307": LOTTOMATICA_ROMA,
}
NBA_VENUE_IDENTITY_MAP: dict[str, str] = {}


class NBACombinedLeagueModel(CombinedLeagueModel):
    """NBA combined implementation of the league model."""

    def __init__(self, session: requests.Session) -> None:
        super().__init__(
            session,
            League.NBA,
            [
                NBANBALeagueModel(session),
                NBAESPNLeagueModel(session),
            ],
        )

    @classmethod
    def team_identity_map(cls) -> dict[str, str]:
        return NBA_TEAM_IDENTITY_MAP

    @classmethod
    def venue_identity_map(cls) -> dict[str, str]:
        return NBA_VENUE_IDENTITY_MAP



"""NBA API team model."""

import pandas as pd

from ....cache import MEMORY
from ...team_model import TeamModel


@MEMORY.cache
def create_nba_nba_team_model(row: pd.Series, home: bool) -> TeamModel | None:
    """Create a team model from NBA API."""
    suffix = "_A" if home else "_B"
    # game_rotation = gamerotation.GameRotation(game_id=row["GAME_ID"])
    # player_dict = {}
    # for _, gr_row in game_rotation.get_data_frames()[int(home)].iterrows():
    #     player_dict[gr_row["PERSON_ID"]] = gr_row
    identifier = row["TEAM_ID" + suffix]
    if identifier is None:
        return None
    return TeamModel(
        identifier=str(identifier),
        name=row["TEAM_NAME" + suffix],
        players=[  # type: ignore
            # create_nba_nba_player_model(v, player_index) for v in player_dict.values()
        ],
        odds=[],
        points=float(row["PTS" + suffix]),
        ladder_rank=None,
        location=None,
    )



"""Combined player model."""

from ..player_model import PlayerModel


def create_combined_player_model(
    player_models: list[PlayerModel], identifier: str
) -> PlayerModel:
    """Create a player model by combining many player models."""
    jersey = None
    kicks = None
    for player_model in player_models:
        player_model_jersey = player_model.jersey
        if player_model_jersey is not None:
            jersey = player_model_jersey
        player_model_kicks = player_model.kicks
        if player_model_kicks is not None:
            kicks = player_model_kicks
    return PlayerModel(identifier=identifier, jersey=jersey, kicks=kicks)



"""The base class for a data model."""

import pandas as pd
import requests


class Model:
    """The base model class."""

    def __init__(self, session: requests.Session) -> None:
        self._session = session

    @property
    def session(self) -> requests.Session:
        """Return the session."""
        return self._session

    def to_frame(self) -> pd.DataFrame:
        """Render the model as a dataframe."""
        raise NotImplementedError("to_frame not implemented by parent class.")



"""The min feature extractor."""

import pandas as pd

from ...data.game_model import (GAME_DT_COLUMN, SEASON_TYPE_COLUMN,
                                SEASON_YEAR_COLUMN)
from ...data.league_model import DELIMITER
from .feature import Feature

MIN_COLUMN_PREFIX = "min"


def _process_season_start_dt(df: pd.DataFrame) -> pd.DataFrame:
    season_start_dt = {}
    season_start_dt_col = DELIMITER.join(["season", "startdt"])
    df[season_start_dt_col] = None

    def record_season_start_dt(row: pd.Series) -> pd.Series:
        nonlocal season_start_dt
        nonlocal season_start_dt_col
        season_key = "-".join([str(row[SEASON_YEAR_COLUMN]), row[SEASON_TYPE_COLUMN]])
        if season_key not in season_start_dt:
            season_start_dt[season_key] = row[GAME_DT_COLUMN]
        row[season_start_dt_col] = season_start_dt[season_key]
        return row

    return df.apply(record_season_start_dt, axis=1)


class MinFeature(Feature):
    """The min feature extractor class."""

    # pylint: disable=too-few-public-methods

    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """Process the dataframe and add the necessary features."""
        df = _process_season_start_dt(df)
        return df.copy()



"""The datetime feature extractor."""

import pandas as pd
from feature_engine.datetime import DatetimeFeatures

from .feature import Feature


class DatetimeFeature(Feature):
    """The datatime feature extractor class."""

    # pylint: disable=too-few-public-methods

    def __init__(self) -> None:
        super().__init__()
        self._dtf = DatetimeFeatures(
            features_to_extract="all", missing_values="ignore", drop_original=False
        )

    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """Process the dataframe and add the necessary features."""
        return self._dtf.fit_transform(df)



"""NFL Sports DB league model."""

from typing import Iterator

import requests

from ...game_model import GameModel
from ...league import League
from ...league_model import LeagueModel
from ...season_type import SeasonType
from .nfl_sportsdb_game_model import create_nfl_sportsdb_game_model


class NFLSportsDBLeagueModel(LeagueModel):
    """NFL SportsDB implementation of the league model."""

    # pylint: disable=too-many-arguments

    def __init__(self, session: requests.Session) -> None:
        super().__init__(League.NFL, session)

    def _produce_games(
        self,
        round_str: str,
        week: int,
        league_id: str,
        season_year: int,
        season_type: SeasonType,
    ) -> Iterator[GameModel]:
        # pylint: disable=line-too-long
        response = self.session.get(
            f"https://www.thesportsdb.com/api/v1/json/3/eventsround.php?id={league_id}&r={round_str}&s={season_year}"
        )
        response.raise_for_status()
        games = response.json()
        events = games["events"]
        if events is None:
            raise ValueError("events is null.")
        for count, game in enumerate(events):
            yield create_nfl_sportsdb_game_model(
                self.session,
                game,
                week,
                count,
                self.league,
                season_year,  # pyright: ignore
                season_type,
            )

    @property
    def games(self) -> Iterator[GameModel]:
        league_id = "4391"
        response = self.session.get(
            f"https://www.thesportsdb.com/api/v1/json/3/search_all_seasons.php?id={league_id}"
        )
        response.raise_for_status()
        seasons = response.json()
        for season in seasons["seasons"]:
            season_year = season["strSeason"]
            for season_type in SeasonType:
                match season_type:
                    case SeasonType.OFFSEASON:
                        return
                    case SeasonType.PRESEASON:
                        try:
                            yield from self._produce_games(
                                str(500), 0, league_id, season_year, season_type
                            )
                        except ValueError:
                            pass
                    case SeasonType.REGULAR:
                        try:
                            for count, round_str in enumerate(range(1, 125)):
                                yield from self._produce_games(
                                    str(round_str),
                                    count,
                                    league_id,
                                    season_year,
                                    season_type,
                                )
                        except ValueError:
                            pass
                    case SeasonType.POSTSEASON:
                        for count, round_str in enumerate(
                            [125, 150, 160, 170, 180, 200]
                        ):
                            try:
                                yield from self._produce_games(
                                    str(round_str),
                                    21 + count,
                                    league_id,
                                    season_year,
                                    season_type,
                                )
                            except ValueError:
                                pass



"""Tests for the class weight class."""
import unittest

import sportsball as spb
import pandas as pd
import numpy as np


class TestClassWeight(unittest.TestCase):

    def setUp(self):
        self._class_weight = spb.strategy.weights.class_weight.ClassWeight()

    def test_process(self):
        y = pd.DataFrame(data={"y": [True, False, True, False, True, True]})
        weight = self._class_weight.process(y)
        np.testing.assert_array_equal(weight, np.array([0.75, 1.5 , 0.75, 1.5 , 0.75, 0.75]))



"""Tests for the skill feature class."""
import datetime
import unittest

import sportsball as spb
import pandas as pd
from sportsball.strategy.features.columns import team_identifier_column, team_points_column, player_identifier_column, player_column_prefix
from sportsball.strategy.features.skill_feature import \
    _find_matches, SkillFeature, SKILL_COLUMN_PREFIX, SKILL_PROBABILITY_COLUMN, YEAR_SLICE_ALL
from sportsball.data.game_model import GAME_DT_COLUMN
from sportsball.data.league_model import DELIMITER
from openskill.models import PlackettLuce


class TestSkillFeatureClass(unittest.TestCase):

    def setUp(self):
        self._skill_feature = SkillFeature(year_slices=[None, 1, 2])

    def test_process(self):
        df = pd.DataFrame(data={
            GAME_DT_COLUMN: [
                datetime.datetime(2009, 12, 31, 0, 0, 0),
                datetime.datetime(2010, 1, 1, 0, 0, 0),
                datetime.datetime(2010, 1, 2, 0, 0, 0),
            ],
            team_identifier_column(0): ["t1", "t1", "t1"],
            team_identifier_column(1): ["t2", "t2", "t2"],
            team_points_column(0): [0, 1, 2],
            team_points_column(1): [3, 4, 5],
            player_identifier_column(0, 0): ["p1", "p1", "p1"],
            player_identifier_column(1, 0): ["p2", "p2", "p2"],
        })
        df = self._skill_feature.process(df)
        # df.to_csv("test_tmp.csv")
        self.assertIsNotNone(df)
        print(df.columns.values)
        print(df)
        probability = df[
            DELIMITER.join([player_column_prefix(0, None), SKILL_COLUMN_PREFIX, "1", SKILL_PROBABILITY_COLUMN])
        ].values.tolist()
        self.assertListEqual(probability, [0.0, 0.34070897206214457, 0.2287564657243515])
        probability = df[
            DELIMITER.join([player_column_prefix(0, None), SKILL_COLUMN_PREFIX, "2", SKILL_PROBABILITY_COLUMN])
        ].values.tolist()
        self.assertListEqual(probability, [0.0, 0.34070897206214457, 0.2287564657243515])
        probability = df[
            DELIMITER.join([player_column_prefix(0, None), SKILL_COLUMN_PREFIX, YEAR_SLICE_ALL, SKILL_PROBABILITY_COLUMN])
        ].values.tolist()
        self.assertListEqual(probability, [0.5, 0.34070897206214457, 0.2287564657243515])


    def test_find_matches(self):
        df = pd.DataFrame(data={
            DELIMITER: [datetime.datetime(2009, 12, 31, 0, 0, 0), datetime.datetime(2010, 1, 1, 0, 0, 0)],
            team_identifier_column(0): ["1", "1"],
            team_identifier_column(1): ["2", "2"],
            team_points_column(0): [0, 1],
            team_points_column(1): [2, 3],
            player_identifier_column(0, 0): ["1", "1"],
            player_identifier_column(1, 0): ["2", "2"],
        })
        team_model = PlackettLuce()
        player_model = PlackettLuce()
        teams = {
            "1": team_model.rating(name="1"),
        }
        players = {
            "1": player_model.rating(name="1"),
        }
        for _, row in df.iterrows():
            _find_matches(row, 1, 5, teams, players)



"""AFL ESPN league model."""

# pylint: disable=line-too-long

import requests

from ...espn.espn_league_model import ESPNLeagueModel
from ...league import League

_SEASON_URL = "https://sports.core.api.espn.com/v2/sports/australian-football/leagues/afl/seasons?limit=100"


class AFLESPNLeagueModel(ESPNLeagueModel):
    """AFL ESPN implementation of the league model."""

    def __init__(self, session: requests.Session) -> None:
        super().__init__(_SEASON_URL, League.AFL, session)



"""A feature reducer combining many sub feature reducers."""

import pandas as pd

from .constant_reducer import ConstantReducer
from .correlation_reducer import CorrelationReducer
from .duplicate_reducer import DuplicateReducer
from .reducer import Reducer


class CombinedReducer(Reducer):
    """Combined reducer extractor class."""

    # pylint: disable=too-few-public-methods

    def __init__(
        self, keep_features: list[str], reducers: list[Reducer] | None = None
    ) -> None:
        super().__init__()
        if reducers is None:
            reducers = [
                ConstantReducer(),
                DuplicateReducer(),
                CorrelationReducer(),
            ]
        self._reducers = reducers
        self._keep_features = keep_features

    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        keep_df = df[self._keep_features].copy()
        for reducer in self._reducers:
            df = reducer.process(df)
        for column in keep_df.columns.values:
            if column not in df.columns.values:
                df[column] = keep_df[column]
        return df



"""The correlation feature reducer."""

import pandas as pd
from feature_engine.selection import DropCorrelatedFeatures

from .reducer import Reducer


class CorrelationReducer(Reducer):
    """The correlation reducer class."""

    # pylint: disable=too-few-public-methods

    def __init__(self) -> None:
        super().__init__()
        self._dcf = DropCorrelatedFeatures(missing_values="ignore")

    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """Process the dataframe and remove the necessary features."""
        return self._dcf.fit_transform(df)



"""Combined team model."""

# pylint: disable=too-many-locals
from ..odds_model import OddsModel
from ..player_model import PlayerModel
from ..team_model import TeamModel
from .combined_player_model import create_combined_player_model


def create_combined_team_model(
    team_models: list[TeamModel],
    identifier: str,
) -> TeamModel:
    """Create a team model by combining many team models."""
    location = None
    players: dict[str, list[PlayerModel]] = {}
    odds: dict[str, list[OddsModel]] = {}
    points = None
    ladder_rank = None
    for team_model in team_models:
        team_model_location = team_model.location
        if team_model_location is not None:
            location = team_model_location
        for player_model in team_model.players:
            key = player_model.jersey
            if key is None:
                key = player_model.identifier
            players[key] = players.get(key, []) + [player_model]
        for odds_model in team_model.odds:
            key = f"{odds_model.bookie.identifier}-{odds_model.odds}"
            odds[key] = odds.get(key, []) + [odds_model]
        team_model_points = team_model.points
        if team_model_points is not None:
            points = team_model_points
        team_model_ladder_rank = team_model.ladder_rank
        if team_model_ladder_rank is not None:
            ladder_rank = team_model_ladder_rank

    return TeamModel(
        identifier=identifier,
        name=team_models[0].name,
        location=location,
        players=[  # pyright: ignore
            create_combined_player_model(x, x[0].identifier) for x in players.values()
        ],
        odds=[x[0] for x in odds.values()],
        points=points,
        ladder_rank=ladder_rank,
    )



"""AFL aussportsbetting game model."""

# pylint: disable=too-many-arguments
import datetime

import requests

from ....cache import MEMORY
from ...game_model import GameModel
from ...league import League
from .afl_aussportsbetting_team_model import \
    create_afl_aussportsbetting_team_model
from .afl_aussportsbetting_venue_model import \
    create_afl_aussportsbetting_venue_model


@MEMORY.cache(ignore=["session"])
def create_afl_aussportsbetting_game_model(
    dt: datetime.datetime,
    home_team: str,
    away_team: str,
    venue: str,
    session: requests.Session,
    home_points: float,
    away_points: float,
    home_odds: float,
    away_odds: float,
    league: League,
) -> GameModel:
    """Create a game model based off aus sports betting."""
    venue_model = create_afl_aussportsbetting_venue_model(venue, session, dt)
    home_team_model = create_afl_aussportsbetting_team_model(
        home_team, home_points, home_odds
    )
    away_team_model = create_afl_aussportsbetting_team_model(
        away_team, away_points, away_odds
    )
    return GameModel(
        dt=dt,
        week=None,
        game_number=None,
        venue=venue_model,
        teams=[home_team_model, away_team_model],  # pyright: ignore
        end_dt=None,
        attendance=None,
        league=league,
        year=None,
        season_type=None,
    )



"""The prototype class for a weight."""

import numpy as np
import pandas as pd


class Weight:
    """The prototype weight class."""

    @classmethod
    def name(cls) -> str:
        """The name of the weight class."""
        raise NotImplementedError("name is not implemented in parent class.")

    def process(self, y: pd.DataFrame) -> np.ndarray:
        """Process the dataframe and remove the necessary features."""
        raise NotImplementedError("process is not implemented in parent class.")



"""ESPN venue model."""

import datetime
from typing import Any

import requests

from ...cache import MEMORY
from ..google.google_address_model import create_google_address_model
from ..venue_model import VenueModel


@MEMORY.cache(ignore=["session"])
def create_espn_venue_model(
    venue: dict[str, Any], session: requests.Session, dt: datetime.datetime
) -> VenueModel:
    """Create a venue model from an ESPN result."""
    identifier = venue["id"]
    name = venue["fullName"]
    venue_address = venue["address"]
    city = venue_address.get("city", "")
    state = venue_address.get("state", "")
    zipcode = venue_address.get("zipCode", "")
    address = create_google_address_model(
        " - ".join([x for x in [name, city, state, zipcode] if x]),
        session,
        dt,
    )
    grass = venue["grass"]
    indoor = venue["indoor"]
    return VenueModel(
        identifier=identifier,
        name=name,
        address=address,  # pyright: ignore
        is_grass=grass,
        is_indoor=indoor,
    )



"""Helper functions for columns."""

import pandas as pd

from ...data.game_model import (GAME_ATTENDANCE_COLUMN, TEAM_COLUMN_PREFIX,
                                VENUE_COLUMN_PREFIX)
from ...data.league_model import DELIMITER
from ...data.player_model import PLAYER_IDENTIFIER_COLUMN, PLAYER_KICKS_COLUMN
from ...data.team_model import (PLAYER_COLUMN_PREFIX, TEAM_IDENTIFIER_COLUMN,
                                TEAM_POINTS_COLUMN)
from ...data.venue_model import VENUE_IDENTIFIER_COLUMN


def team_column_prefix(team_idx: int) -> str:
    """Generate a prefix for a team column at a given index."""
    return DELIMITER.join(
        [
            TEAM_COLUMN_PREFIX,
            str(team_idx),
        ]
    )


def team_identifier_column(team_idx: int) -> str:
    """Generate a team identifier column at a given index."""
    return DELIMITER.join([team_column_prefix(team_idx), TEAM_IDENTIFIER_COLUMN])


def team_points_column(team_idx: int) -> str:
    """Generate a team points column at a given index."""
    return DELIMITER.join([team_column_prefix(team_idx), TEAM_POINTS_COLUMN])


def player_column_prefix(team_idx: int, player_idx: int | None) -> str:
    """Generate a prefix for a player column at a given index."""
    if player_idx is None:
        return DELIMITER.join(
            [
                team_column_prefix(team_idx),
                PLAYER_COLUMN_PREFIX,
            ]
        )
    return DELIMITER.join(
        [
            team_column_prefix(team_idx),
            PLAYER_COLUMN_PREFIX,
            str(player_idx),
        ]
    )


def player_identifier_column(team_idx: int, player_idx: int) -> str:
    """Generate a team points column at a given index."""
    return DELIMITER.join(
        [player_column_prefix(team_idx, player_idx), PLAYER_IDENTIFIER_COLUMN]
    )


def attendance_column() -> str:
    """Generate an attendance column."""
    return DELIMITER.join([GAME_ATTENDANCE_COLUMN])


def find_team_count(df: pd.DataFrame) -> int:
    """Find the number of teams in the dataframe."""
    team_count = 0
    while True:
        if team_identifier_column(team_count) not in df.columns.values:
            break
        team_count += 1
    return team_count


def find_player_count(df: pd.DataFrame, team_count: int) -> int:
    """Find the number of players in a team in the dataframe."""
    player_count = 0
    while True:
        found_player = False
        for i in range(team_count):
            if player_identifier_column(i, player_count) not in df.columns.values:
                continue
            found_player = True
        if not found_player:
            break
        player_count += 1
    return player_count


def venue_identifier_column() -> str:
    """Generate a venue identifier column."""
    return DELIMITER.join([VENUE_COLUMN_PREFIX, VENUE_IDENTIFIER_COLUMN])


def kick_column(team_idx: int, player_idx: int) -> str:
    """Generate a kick column."""
    return DELIMITER.join(
        [player_column_prefix(team_idx, player_idx), PLAYER_KICKS_COLUMN]
    )



"""The prototype class for a team."""

from typing import Any, Literal

from pydantic import BaseModel, Field

from .field_type import TYPE_KEY, FieldType
from .odds_model import OddsModel
from .player_model import PlayerModel

TEAM_POINTS_COLUMN: Literal["points"] = "points"
TEAM_IDENTIFIER_COLUMN: Literal["identifier"] = "identifier"
PLAYER_COLUMN_PREFIX: Literal["players"] = "players"


def _calculate_kicks(data: dict[str, Any]) -> int | None:
    kicks = 0
    found_kicks = False
    for player in data[PLAYER_COLUMN_PREFIX]:
        player_kicks = player.kicks
        if player_kicks is None:
            continue
        found_kicks = True
        kicks += player_kicks
    if not found_kicks:
        return None
    return kicks


class TeamModel(BaseModel):
    """The serialisable team class."""

    identifier: str = Field(
        ...,
        json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL},
        alias=TEAM_IDENTIFIER_COLUMN,
    )
    name: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.TEXT})
    location: str | None = Field(
        ..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL}
    )
    players: list[PlayerModel] = Field(..., alias=PLAYER_COLUMN_PREFIX)
    odds: list[OddsModel]
    points: float | None = Field(
        ..., json_schema_extra={TYPE_KEY: FieldType.POINTS}, alias=TEAM_POINTS_COLUMN
    )
    ladder_rank: int | None
    kicks: int | None = Field(
        default_factory=_calculate_kicks,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
    )



"""The vennabers class for a trainer."""

import os
from typing import Any

import joblib  # type: ignore
import pandas as pd
from venn_abers import VennAbers  # type: ignore

from .output_column import output_prob_column
from .trainer import Trainer

_MODEL_FILENAME = "va.sav"


class VennAbersTrainer(Trainer):
    """The catboost trainer class."""

    def __init__(self, folder: str, wrapped_trainer: Trainer) -> None:
        super().__init__(folder)
        self._wrapped_trainer = wrapped_trainer
        self._model = VennAbers()

    @property
    def clf(self) -> Any:
        """The underlying classifier"""
        return self._model

    @property
    def salt(self) -> str:
        """The salt to use when hashing the predictions."""
        return "vennabers-" + self._wrapped_trainer.salt

    def fit(
        self,
        x: tuple[pd.DataFrame, pd.DataFrame | None],
        y: tuple[pd.DataFrame, pd.DataFrame | None],
    ):
        """Fit the data."""
        self._wrapped_trainer.fit(x, y)
        y_pred = self._wrapped_trainer.predict_proba(x[0])
        if y_pred is None:
            raise ValueError("y_pred is null")
        self._model.fit(y_pred.to_numpy(), y[0].to_numpy())

    def save(self):
        """Save the trainer."""
        self._wrapped_trainer.save()
        joblib.dump(self._model, os.path.join(self._folder, _MODEL_FILENAME))

    def load(self):
        """Load the trainer."""
        self._wrapped_trainer.load()
        self._model = joblib.load(os.path.join(self._folder, _MODEL_FILENAME))

    def predict(self, x: pd.DataFrame) -> pd.DataFrame | None:
        """Predict the Y values."""
        return self._wrapped_trainer.predict(x)

    def predict_proba(self, x: pd.DataFrame) -> pd.DataFrame | None:
        """Predict the Y probabilities."""
        y = super().predict(x)
        if y is not None:
            return y

        y_prob = self._wrapped_trainer.predict_proba(x)
        if y_prob is None:
            raise ValueError("y_prob is null")

        p_prime, _ = self._model.predict_proba(y_prob.to_numpy())
        y = pd.DataFrame(
            index=x.index,
            data={
                output_prob_column(i): p_prime[:, i]
                for i in range(len(y_prob.columns.values))
            },
        )

        self.save_prediction_proba(x, y)
        return y

    def select_features(
        self,
        x: tuple[pd.DataFrame, pd.DataFrame | None],
        y: tuple[pd.DataFrame, pd.DataFrame | None],
    ) -> tuple[list[str], int]:
        """Select the features from the training data."""
        return self._wrapped_trainer.select_features(x, y)



"""The prototype class for a feature."""

import pandas as pd


class Feature:
    """The prototype feature class."""

    # pylint: disable=too-few-public-methods

    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """Process the dataframe and add the necessary features."""
        raise NotImplementedError("process is not implemented in parent class.")



"""A weight based on sigmoid bias towards the current time."""

import numpy as np
import pandas as pd
from scipy.special import expit  # type: ignore

from .weight import Weight


class SigmoidWeight(Weight):
    """Sigmoid weight class."""

    @classmethod
    def name(cls) -> str:
        """The name of the weight class."""
        return "sigmoid"

    def process(self, y: pd.DataFrame) -> np.ndarray:
        return expit(np.linspace(0, 1.0, len(y)))



"""A class for holding address information."""

from pydantic import BaseModel, Field

from .field_type import TYPE_KEY, FieldType
from .weather_model import WeatherModel


class AddressModel(BaseModel):
    """The class for representing an address."""

    city: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})
    state: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})
    zipcode: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})
    latitude: float | None
    longitude: float | None
    housenumber: str | None = Field(
        ..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL}
    )
    weather: WeatherModel | None
    timezone: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})



"""Tests for the lag feature class."""
import unittest

import pandas as pd
import numpy as np
from sportsball.strategy.features.columns import team_identifier_column, venue_identifier_column, attendance_column, kick_column, player_identifier_column
from sportsball.strategy.features.lag_feature import LagFeature, LAG_COLUMN_PREFIX
from sportsball.data.league_model import DELIMITER


class TestLagFeature(unittest.TestCase):

    def setUp(self):
        self._lag_feature = LagFeature()

    def test_process(self):
        df = pd.DataFrame(data={
            team_identifier_column(0): ["1", "1"],
            team_identifier_column(1): ["2", "2"],
            venue_identifier_column(): ["4", "4"],
            attendance_column(): [12000, 13000]
        })
        df = self._lag_feature.process(df)
        lag_attendances = df[DELIMITER.join([LAG_COLUMN_PREFIX, attendance_column()])].values.tolist()[1:]
        self.assertListEqual(lag_attendances, [12000.0])

    def test_kicks(self):
        df = pd.DataFrame(data={
            team_identifier_column(0): ["1", "1"],
            team_identifier_column(1): ["2", "2"],
            venue_identifier_column(): ["4", "4"],
            attendance_column(): [12000, 13000],
            player_identifier_column(0, 0): ["p1", "p1"],
            player_identifier_column(0, 1): ["p2", "p2"],
            player_identifier_column(1, 0): ["p3", "p3"],
            player_identifier_column(1, 1): ["p4", "p4"],
            kick_column(0, 0): [10, 20],
            kick_column(0, 1): [20, 30],
            kick_column(1, 0): [30, 40],
            kick_column(1, 1): [40, 50],
        })
        df = self._lag_feature.process(df)
        lag_kicks = df[DELIMITER.join([LAG_COLUMN_PREFIX, kick_column(0, 0)])].values.tolist()[1:]
        self.assertListEqual(lag_kicks, [10.0])
        lag_kicks = df[DELIMITER.join([LAG_COLUMN_PREFIX, kick_column(0, 1)])].values.tolist()[1:]
        self.assertListEqual(lag_kicks, [20.0])



"""Google address model."""

# pylint: disable=too-many-lines,line-too-long
import datetime
import logging
from collections import namedtuple
from typing import Any

import geocoder  # type: ignore
import requests
from timezonefinder import TimezoneFinder  # type: ignore

from ...cache import MEMORY
from ..address_model import AddressModel
from ..openmeteo.openmeteo_weather_model import create_openmeteo_weather_model

SportsballGeocodeTuple = namedtuple(
    "SportsballGeocodeTuple", ["city", "state", "postal", "lat", "lng", "housenumber"]
)
LEVI_STADIUM = SportsballGeocodeTuple(
    city="Santa Clara",
    state="CA",
    postal="95054",
    lat=37.4033165,
    lng=-121.9719523,
    housenumber="4900",
)
ARROWHEAD_STADIUM = SportsballGeocodeTuple(
    city="Kansas City",
    state="MO",
    postal="64129",
    lat=39.0489647,
    lng=-94.486589,
    housenumber="1",
)
PAYCOR_STADIUM = SportsballGeocodeTuple(
    city="Cincinnati",
    state="OH",
    postal="45202",
    lat=39.0954576,
    lng=-84.5186326,
    housenumber="1",
)
LUMEN_FIELD = SportsballGeocodeTuple(
    city="Seattle",
    state="WA",
    postal="98134",
    lat=47.5951518,
    lng=-122.3342143,
    housenumber="800",
)
CAESARS_SUPERDOME = SportsballGeocodeTuple(
    city="New Orleans",
    state="LA",
    postal="70112",
    lat=29.951061,
    lng=-90.0838191,
    housenumber="1500",
)
TOM_BENSON_HALL_OF_FAME_STADIUM = SportsballGeocodeTuple(
    city="Canton",
    state="OH",
    postal="44708",
    lat=40.8198713,
    lng=-81.4028911,
    housenumber="",
)
GILETTE_STADIUM = SportsballGeocodeTuple(
    city="Foxborough",
    state="MA",
    postal="02035",
    lat=42.0909458,
    lng=-71.2669214,
    housenumber="1",
)
FORD_FIELD = SportsballGeocodeTuple(
    city="Detroit",
    state="MI",
    postal="48226",
    lat=42.3400064,
    lng=-83.0481779,
    housenumber="2000",
)
HARD_ROCK_STADIUM = SportsballGeocodeTuple(
    city="Miami Gardens",
    state="FL",
    postal="33056",
    lat=25.958326,
    lng=-80.2421728,
    housenumber="100",
)
RAYMOND_JAMES_STADIUM = SportsballGeocodeTuple(
    city="Tampa",
    state="FL",
    postal="33607",
    lat=27.9758691,
    lng=-82.5059093,
    housenumber="4201",
)
CLEVELAND_BROWNS_STADIUM = SportsballGeocodeTuple(
    city="Cleveland",
    state="OH",
    postal="44114",
    lat=41.5060535,
    lng=-81.702123,
    housenumber="100",
)
STATE_FARM_STADIUM = SportsballGeocodeTuple(
    city="Glendale",
    state="AZ",
    postal="85305",
    lat=33.5276247,
    lng=-112.2651342,
    housenumber="1001",
)
HIGHMARK_STADIUM = SportsballGeocodeTuple(
    city="Orchard Park",
    state="NY",
    postal="14127",
    lat=42.7737546,
    lng=-78.7895472,
    housenumber="1",
)
SOLDIER_FIELD = SportsballGeocodeTuple(
    city="Chicago",
    state="IL",
    postal="60605",
    lat=41.8623132,
    lng=-87.6192633,
    housenumber="1410",
)
BANK_OF_AMERICA_STADIUM = SportsballGeocodeTuple(
    city="Charlotte",
    state="NC",
    postal="28202",
    lat=35.2257795,
    lng=-80.8555399,
    housenumber="800",
)
ATT_STADIUM = SportsballGeocodeTuple(
    city="Arlington",
    state="TX",
    postal="76011",
    lat=32.7479966,
    lng=-97.0959914,
    housenumber="1",
)
MTT_BANK_STADIUM = SportsballGeocodeTuple(
    city="Baltimore",
    state="FL",
    postal="21230",
    lat=39.2778666,
    lng=-76.6255186,
    housenumber="101",
)
SOFI_STADIUM = SportsballGeocodeTuple(
    city="Inglewood",
    state="CA",
    postal="90301",
    lat=33.9534651,
    lng=-118.3416145,
    housenumber="1001",
)
ALLEGIANT_STADIUM = SportsballGeocodeTuple(
    city="Las Vegas",
    state="NV",
    postal="89118",
    lat=36.0908665,
    lng=-115.1858971,
    housenumber="3333",
)
LINCOLN_FINANCIAL_FIELD = SportsballGeocodeTuple(
    city="Philadelphia",
    state="PA",
    postal="19148",
    lat=39.9013695,
    lng=-75.1700964,
    housenumber="",
)
METLIFE_STADIUM = SportsballGeocodeTuple(
    city="East Rutherford",
    state="NJ",
    postal="07073",
    lat=40.8135064,
    lng=-74.0770319,
    housenumber="1",
)
MERCEDES_BANZ_STADIUM = SportsballGeocodeTuple(
    city="Atlanta",
    state="GA",
    postal="30313",
    lat=33.7553232,
    lng=-84.4031654,
    housenumber="1",
)
NRG_STADIUM = SportsballGeocodeTuple(
    city="Houston",
    state="TX",
    postal="77054",
    lat=29.6847219,
    lng=-95.4132823,
    housenumber="",
)
ACRISURE_STADIUM = SportsballGeocodeTuple(
    city="Pittsburgh",
    state="PA",
    postal="15212",
    lat=40.4467648,
    lng=-80.0183352,
    housenumber="100",
)
LUCAS_OIL_STADIUM = SportsballGeocodeTuple(
    city="Indianapolis",
    state="IN",
    postal="46225",
    lat=39.7601007,
    lng=-86.1664626,
    housenumber="500",
)
LAMBEAU_FIELD = SportsballGeocodeTuple(
    city="Green Bay",
    state="WI",
    postal="54304",
    lat=44.5013406,
    lng=-88.0647832,
    housenumber="1265",
)
US_BANK_STADIUM = SportsballGeocodeTuple(
    city="Minneapolis",
    state="MN",
    postal="55415",
    lat=44.9736461,
    lng=-93.2600694,
    housenumber="401",
)
FEDEX_FIELD = SportsballGeocodeTuple(
    city="Landover",
    state="MD",
    postal="20785",
    lat=38.9081324,
    lng=-76.8665633,
    housenumber="1600",
)
NISSAN_STADIUM = SportsballGeocodeTuple(
    city="Nashville",
    state="TN",
    postal="37213",
    lat=36.166469,
    lng=-86.7738744,
    housenumber="1",
)
EVERBANK_STADIUM = SportsballGeocodeTuple(
    city="Jacksonville",
    state="FL",
    postal="32202",
    lat=30.3239099,
    lng=-81.6399246,
    housenumber="1",
)
EMPOWER_FIELD = SportsballGeocodeTuple(
    city="Denver",
    state="CO",
    postal="80204",
    lat=39.7438895,
    lng=-105.0226843,
    housenumber="1701",
)
WEMBLEY_STADIUM = SportsballGeocodeTuple(
    city="Wembley",
    state="",
    postal="HA9 0WS",
    lat=51.5560247,
    lng=-0.2821926,
    housenumber="",
)
TOTTENHAM_HOTSPURS_STADIUM = SportsballGeocodeTuple(
    city="London",
    state="",
    postal="N17 0BX",
    lat=51.6042204,
    lng=-0.071094,
    housenumber="782",
)
DEUTSCHE_BANK_PARK = SportsballGeocodeTuple(
    city="Frankfurt am Main",
    state="",
    postal="60528",
    lat=50.0685807,
    lng=8.6428971,
    housenumber="362",
)
ALLIANZ_ARENA = SportsballGeocodeTuple(
    city="München",
    state="",
    postal="80939",
    lat=48.2187966,
    lng=11.6221325,
    housenumber="25",
)
GEORGIA_DOME = SportsballGeocodeTuple(
    city="Atlanta",
    state="GA",
    postal="",
    lat=33.758,
    lng=-84.401,
    housenumber="1",
)
TRANS_WORLD_DOME = SportsballGeocodeTuple(
    city="St. Louis",
    state="MO",
    postal="63101",
    lat=38.6328232,
    lng=-90.1910817,
    housenumber="701",
)
CANDLESTICK_PARK = SportsballGeocodeTuple(
    city="San Francisco",
    state="CA",
    postal="94124",
    lat=37.7129826,
    lng=-122.4064101,
    housenumber="",
)
SILVERDOME = SportsballGeocodeTuple(
    city="Pontiac",
    state="MI",
    postal="48342",
    lat=42.6456689,
    lng=-83.2572144,
    housenumber="",
)
TEXAS_STADIUM = SportsballGeocodeTuple(
    city="Irving",
    state="TX",
    postal="",
    lat=32.839722,
    lng=-96.9146954,
    housenumber="",
)
SDCCU_STADIUM = SportsballGeocodeTuple(
    city="San Diego",
    state="CA",
    postal="92108",
    lat=32.7809825,
    lng=-117.1244523,
    housenumber="9449",
)
MOUNTAIN_AMERICA_STADIUM = SportsballGeocodeTuple(
    city="Tempe",
    state="AZ",
    postal="85287",
    lat=33.4264574,
    lng=-111.9374442,
    housenumber="500",
)
LOS_ANGELES_MEMORIAL_COLESIUM = SportsballGeocodeTuple(
    city="Los Angeles",
    state="CA",
    postal="90037",
    lat=34.0136647,
    lng=-118.2904104,
    housenumber="3911",
)
KEZAR_STADIUM = SportsballGeocodeTuple(
    city="San Francisco",
    state="CA",
    postal="94118",
    lat=37.7669182,
    lng=-122.4567445,
    housenumber="670",
)
CITY_STADIUM = SportsballGeocodeTuple(
    city="Green Bay",
    state="WI",
    postal="54301",
    lat=44.5076658,
    lng=-87.9975486,
    housenumber="1500",
)
COTTON_BOWL = SportsballGeocodeTuple(
    city="Dallas",
    state="TX",
    postal="75215",
    lat=32.7796106,
    lng=-96.7602328,
    housenumber="3750",
)
WRIGLEY_FIELD = SportsballGeocodeTuple(
    city="Chicago",
    state="IL",
    postal="60613",
    lat=41.9484384,
    lng=-87.6579076,
    housenumber="1060",
)
YANKEE_STADIUM = SportsballGeocodeTuple(
    city="Bronx",
    state="NY",
    postal="10451",
    lat=40.8296425,
    lng=-73.9310454,
    housenumber="1",
)
TIGER_STADIUM = SportsballGeocodeTuple(
    city="Baton Rouge",
    state="LA",
    postal="70803",
    lat=30.412035,
    lng=-91.1863912,
    housenumber="",
)
RFK_STADIUM = SportsballGeocodeTuple(
    city="Washington",
    state="DC",
    postal="20003",
    lat=38.8898525,
    lng=-76.9731573,
    housenumber="2400",
)
WAR_MEMORIAL_STADIUM = SportsballGeocodeTuple(
    city="Little Rock",
    state="AR",
    postal="72205",
    lat=34.749861,
    lng=-92.3314143,
    housenumber="",
)
SHEA_STADIUM = SportsballGeocodeTuple(
    city="Flushing",
    state="NY",
    postal="11368",
    lat=40.7570861,
    lng=-73.8653052,
    housenumber="41",
)
MUNICIPAL_STADIUM = SportsballGeocodeTuple(
    city="Kansas City",
    state="MS",
    postal="",
    lat=39.087,
    lng=-94.558,
    housenumber="2123",
)
YALE_BOWL = SportsballGeocodeTuple(
    city="New Haven",
    state="CT",
    postal="06515",
    lat=41.3132191,
    lng=-72.9630891,
    housenumber="81",
)
RICE_STADIUM = SportsballGeocodeTuple(
    city="Houston",
    state="TX",
    postal="77005",
    lat=29.7162887,
    lng=-95.41188622,
    housenumber="6100",
)
ROSE_BOWL = SportsballGeocodeTuple(
    city="Pasadena",
    state="CA",
    postal="91103",
    lat=34.1613284,
    lng=-118.1702211,
    housenumber="1001",
)
MILWAUKEE_COUNTY_STADIUM = SportsballGeocodeTuple(
    city="Milwaukee",
    state="WI",
    postal="53208",
    lat=43.0303919,
    lng=-87.9787396,
    housenumber="201",
)
STANFORD_STADIUM = SportsballGeocodeTuple(
    city="Stanford",
    state="CA",
    postal="94305",
    lat=37.4345298,
    lng=-122.1636976,
    housenumber="625",
)
LIBERTY_BOWL_MEMORIAL_STADIUM = SportsballGeocodeTuple(
    city="Memphis",
    state="TN",
    postal="38119",
    lat=35.1213707,
    lng=-89.9969239,
    housenumber="959",
)
NEO_QUIMICA_ARENA = SportsballGeocodeTuple(
    city="São Paulo",
    state="SP",
    postal="08295-005",
    lat=-23.5453134,
    lng=-46.4768041,
    housenumber="111",
)
MCG = SportsballGeocodeTuple(
    city="Richmond",
    state="VIC",
    postal="3002",
    lat=-37.8199668,
    lng=144.9785784,
    housenumber="",
)
SCG = SportsballGeocodeTuple(
    city="Moore Park",
    state="NSW",
    postal="2021",
    lat=-33.8915316,
    lng=151.2248515,
    housenumber="",
)
ENGIE_STADIUM = SportsballGeocodeTuple(
    city="Sydney Olympic Park",
    state="NSW",
    postal="2127",
    lat=-33.8471156,
    lng=151.0608386,
    housenumber="",
)
ADELAIDE_OVAL = SportsballGeocodeTuple(
    city="North Adelaide",
    state="SA",
    postal="5006",
    lat=-34.9156273,
    lng=138.5914356,
    housenumber="70",
)
GABBA = SportsballGeocodeTuple(
    city="Woolloongabba",
    state="QLD",
    postal="4102",
    lat=-27.4858375,
    lng=153.0332144,
    housenumber="",
)
OPTUS_STADIUM = SportsballGeocodeTuple(
    city="Burswood",
    state="WA",
    postal="6100",
    lat=-31.9511597,
    lng=115.884175,
    housenumber="333",
)
MARVEL_STADIUM = SportsballGeocodeTuple(
    city="Docklands",
    state="VIC",
    postal="3008",
    lat=-37.8165647,
    lng=144.9449306,
    housenumber="740",
)
MARS_STADIUM = SportsballGeocodeTuple(
    city="Wendouree",
    state="VIC",
    postal="3355",
    lat=-37.5386166,
    lng=143.8273839,
    housenumber="725",
)
GMHBA_STADIUM = SportsballGeocodeTuple(
    city="South Geelong",
    state="VIC",
    postal="3220",
    lat=-38.1579979,
    lng=144.3520981,
    housenumber="370",
)
UTAS_STADIUM = SportsballGeocodeTuple(
    city="Invermay",
    state="TAS",
    postal="7248",
    lat=-41.4259371,
    lng=147.137879,
    housenumber="2",
)
PEOPLE_FIRST_STADIUM = SportsballGeocodeTuple(
    city="Carrara",
    state="QLD",
    postal="4211",
    lat=-28.0063039,
    lng=153.3645259,
    housenumber="",
)
BLUNDSTONE_ARENA = SportsballGeocodeTuple(
    city="Bellerive",
    state="TAS",
    postal="7018",
    lat=-42.8772962,
    lng=147.3691419,
    housenumber="15",
)
MANUKA_OVAL = SportsballGeocodeTuple(
    city="Griffith",
    state="ACT",
    postal="2603",
    lat=-35.3181329,
    lng=149.1297911,
    housenumber="",
)
TIO_STADIUM = SportsballGeocodeTuple(
    city="Marrara",
    state="NT",
    postal="0812",
    lat=-12.3991939,
    lng=130.8847367,
    housenumber="70",
)
NORWOOD_OVAL = SportsballGeocodeTuple(
    city="Norwood",
    state="SA",
    postal="5067",
    lat=-34.9198089,
    lng=138.6278735,
    housenumber="",
)
ADELAIDE_HILLS = SportsballGeocodeTuple(
    city="Mount Barker Summit",
    state="SA",
    postal="5251",
    lat=-35.0779754,
    lng=138.8908443,
    housenumber="304",
)
TRAEGER_PARK = SportsballGeocodeTuple(
    city="The Gap",
    state="NT",
    postal="0870",
    lat=-23.7090566,
    lng=133.8701488,
    housenumber="",
)
CAZALYS_STADIUM = SportsballGeocodeTuple(
    city="Westcourt",
    state="QLD",
    postal="4870",
    lat=-16.9357343,
    lng=145.7464957,
    housenumber="",
)
RIVERWAY_STADIUM = SportsballGeocodeTuple(
    city="Thuringowa Central",
    state="QLD",
    postal="4817",
    lat=-19.3176321,
    lng=146.7270047,
    housenumber="",
)
JIANGWAN_STADIUM = SportsballGeocodeTuple(
    city="Shanghai",
    state="",
    postal="200433",
    lat=31.3076026,
    lng=121.5160351,
    housenumber="",
)
DOMAIN_STADIUM = SportsballGeocodeTuple(
    city="Subiaco",
    state="WA",
    postal="6008",
    lat=-31.9444594,
    lng=115.8285952,
    housenumber="304",
)
WESTPAC_STADIUM = SportsballGeocodeTuple(
    city="Pipitea",
    state="",
    postal="6140",
    lat=-41.2729198,
    lng=174.7846279,
    housenumber="105",
)
AAMI_STADIUM = SportsballGeocodeTuple(
    city="West Lakes",
    state="SA",
    postal="5021",
    lat=-34.8797138,
    lng=138.4910204,
    housenumber="",
)
BLACKTOWN_PARK = SportsballGeocodeTuple(
    city="Rooty Hill",
    state="NSW",
    postal="2766",
    lat=-33.7695492,
    lng=150.8558658,
    housenumber="",
)
_CACHED_GEOCODES: dict[str, Any] = {
    "S.C.G. - Australia": SCG,
    "Victoria Park - Australia": SportsballGeocodeTuple(
        city="Abbotsford",
        state="VIC",
        postal="3067",
        lat=-37.8105167,
        lng=144.9822987,
        housenumber="",
    ),
    "Tom Benson Hall of Fame Stadium": TOM_BENSON_HALL_OF_FAME_STADIUM,
    "MetLife Stadium": METLIFE_STADIUM,
    "Gillette Stadium": GILETTE_STADIUM,
    "Acrisure Stadium": ACRISURE_STADIUM,
    "Hard Rock Stadium": HARD_ROCK_STADIUM,
    "M&T Bank Stadium": MTT_BANK_STADIUM,
    "Highmark Stadium": HIGHMARK_STADIUM,
    "U.S. Bank Stadium": US_BANK_STADIUM,
    "Cleveland Browns Stadium": CLEVELAND_BROWNS_STADIUM,
    "EverBank Stadium": EVERBANK_STADIUM,
    "Nissan Stadium": NISSAN_STADIUM,
    "Paycor Stadium": PAYCOR_STADIUM,
    "SoFi Stadium": SOFI_STADIUM,
    "State Farm Stadium": STATE_FARM_STADIUM,
    "Lucas Oil Stadium": LUCAS_OIL_STADIUM,
    "NRG Stadium": NRG_STADIUM,
    "Soldier Field": SOLDIER_FIELD,
    "GEHA Field at Arrowhead Stadium": ARROWHEAD_STADIUM,
    "Bank of America Stadium": BANK_OF_AMERICA_STADIUM,
    "Allegiant Stadium": ALLEGIANT_STADIUM,
    "Empower Field at Mile High": EMPOWER_FIELD,
    "Levi's Stadium": LEVI_STADIUM,
    "Mercedes-Benz Stadium": MERCEDES_BANZ_STADIUM,
    "Raymond James Stadium": RAYMOND_JAMES_STADIUM,
    "Lambeau Field": LAMBEAU_FIELD,
    "Lincoln Financial Field": LINCOLN_FINANCIAL_FIELD,
    "Ford Field": FORD_FIELD,
    "AT&T Stadium": ATT_STADIUM,
    "Lumen Field": LUMEN_FIELD,
    "Caesars Superdome": CAESARS_SUPERDOME,
    "Neo Química Arena": NEO_QUIMICA_ARENA,
    "FedExField": FEDEX_FIELD,
    "Tottenham Hotspur Stadium": TOTTENHAM_HOTSPURS_STADIUM,
    "Wembley Stadium": WEMBLEY_STADIUM,
    "Allianz Arena": ALLIANZ_ARENA,
    "Levi Stadium": LEVI_STADIUM,
    "Levi’s Stadium": LEVI_STADIUM,
    "Deutsche Bank Park": DEUTSCHE_BANK_PARK,
    "Arrowhead Stadium": ARROWHEAD_STADIUM,
    "Paul Brown Stadium": PAYCOR_STADIUM,
    "Trans World Dome": TRANS_WORLD_DOME,
    "Kingdome": LUMEN_FIELD,
    "Georgia Dome": GEORGIA_DOME,
    "Louisiana Superdome": CAESARS_SUPERDOME,
    "Tom Benson Hall of Fame Stadium - Canton - OH - 44708": TOM_BENSON_HALL_OF_FAME_STADIUM,
    "Gillette Stadium - Foxborough - MA - 02035": GILETTE_STADIUM,
    "Lumen Field - Seattle - WA - 98134": LUMEN_FIELD,
    "Paycor Stadium - Cincinnati - OH - 45202": PAYCOR_STADIUM,
    "Ford Field - Detroit - MI - 48226": FORD_FIELD,
    "Hard Rock Stadium - Miami Gardens - FL - 33056": HARD_ROCK_STADIUM,
    "Raymond James Stadium - Tampa - FL - 33607": RAYMOND_JAMES_STADIUM,
    "Huntington Bank Field - Cleveland - OH - 44114": CLEVELAND_BROWNS_STADIUM,
    "State Farm Stadium - Glendale - AZ - 85305": STATE_FARM_STADIUM,
    "Highmark Stadium - Orchard Park - NY - 14127": HIGHMARK_STADIUM,
    "Soldier Field - Chicago - IL - 60605": SOLDIER_FIELD,
    "Bank of America Stadium - Charlotte - NC - 28202": BANK_OF_AMERICA_STADIUM,
    "AT&T Stadium - Arlington - TX - 76011": ATT_STADIUM,
    "M&T Bank Stadium - Baltimore - MD - 21230": MTT_BANK_STADIUM,
    "SoFi Stadium - Inglewood - CA - 90301": SOFI_STADIUM,
    "Caesars Superdome - New Orleans - LA - 70112": CAESARS_SUPERDOME,
    "Allegiant Stadium - Las Vegas - NV - 89118": ALLEGIANT_STADIUM,
    "Lincoln Financial Field - Philadelphia - PA - 19148": LINCOLN_FINANCIAL_FIELD,
    "MetLife Stadium - East Rutherford - NJ - 07073": METLIFE_STADIUM,
    "Mercedes-Benz Stadium - Atlanta - GA - 30313": MERCEDES_BANZ_STADIUM,
    "NRG Stadium - Houston - TX - 77054": NRG_STADIUM,
    "Acrisure Stadium - Pittsburgh - PA - 15212": ACRISURE_STADIUM,
    "Lucas Oil Stadium - Indianapolis - IN - 46225": LUCAS_OIL_STADIUM,
    "Lambeau Field - Green Bay - WI - 54304": LAMBEAU_FIELD,
    "U.S. Bank Stadium - Minneapolis - MN - 55415": US_BANK_STADIUM,
    "Levi's Stadium - Santa Clara - CA - 95054": LEVI_STADIUM,
    "Northwest Stadium - Landover - MD - 20785": FEDEX_FIELD,
    "Nissan Stadium - Nashville - TN - 37213": NISSAN_STADIUM,
    "GEHA Field at Arrowhead Stadium - Kansas City - MO - 64129": ARROWHEAD_STADIUM,
    "EverBank Stadium - Jacksonville - FL - 32202": EVERBANK_STADIUM,
    "Empower Field at Mile High - Denver - CO - 80204": EMPOWER_FIELD,
    "Brunswick St - Australia": SportsballGeocodeTuple(
        city="Fitzroy North",
        state="VIC",
        postal="3068",
        lat=-37.7891016,
        lng=144.9759327,
        housenumber="1",
    ),
    "Corio Oval - Australia": SportsballGeocodeTuple(
        city="East Geelong",
        state="VIC",
        postal="3219",
        lat=-38.1525822,
        lng=144.3787522,
        housenumber="5",
    ),
    "Lake Oval - Australia": SportsballGeocodeTuple(
        city="Albert Park",
        state="VIC",
        postal="3206",
        lat=-37.8403445,
        lng=144.9458567,
        housenumber="",
    ),
    "East Melbourne - Australia": SportsballGeocodeTuple(
        city="Richmond",
        state="VIC",
        postal="3002",
        lat=-37.8174763,
        lng=144.9777351,
        housenumber="",
    ),
    "Junction Oval - Australia": SportsballGeocodeTuple(
        city="St Kilda",
        state="VIC",
        postal="3182",
        lat=-37.8559698,
        lng=144.9779958,
        housenumber="",
    ),
    "M.C.G. - Australia": MCG,
    "Princes Park - Australia": SportsballGeocodeTuple(
        city="Carlton North",
        state="VIC",
        postal="3054",
        lat=-37.7849821,
        lng=144.9565047,
        housenumber="200-590",
    ),
    "Wembley Stadium - London": WEMBLEY_STADIUM,
    "Tottenham Hotspur Stadium - London": TOTTENHAM_HOTSPURS_STADIUM,
    "Punt Rd - Australia": SportsballGeocodeTuple(
        city="East Melbourne",
        state="VIC",
        postal="3002",
        lat=-37.8222208,
        lng=144.9856865,
        housenumber="",
    ),
    "Frankfurt Stadium - Frankfurt": DEUTSCHE_BANK_PARK,
    "Windy Hill - Australia": SportsballGeocodeTuple(
        city="Essendon",
        state="VIC",
        postal="3040",
        lat=-37.7515878,
        lng=144.9001347,
        housenumber="",
    ),
    "Camping World Stadium - Orlando - FL - 32805": SportsballGeocodeTuple(
        city="Orlando",
        state="FL",
        postal="32805",
        lat=28.5390192,
        lng=-81.4033962,
        housenumber="1",
    ),
    "Allianz Arena - Munich": ALLIANZ_ARENA,
    "Estadio Azteca - Mexico City": SportsballGeocodeTuple(
        city="Ciudad de México",
        state="CDMX",
        postal="04650",
        lat=19.3028606,
        lng=-99.1553986,
        housenumber="3465",
    ),
    "Glenferrie Oval - Australia": SportsballGeocodeTuple(
        city="Hawthorn",
        state="VIC",
        postal="3122",
        lat=-37.8206183,
        lng=145.0303644,
        housenumber="34",
    ),
    "Dignity Health Sports Park - Carson - CA - 90746": SportsballGeocodeTuple(
        city="Carson",
        state="CA",
        postal="90746",
        lat=33.8643776,
        lng=-118.2660135,
        housenumber="18400",
    ),
    "Arden St - Australia": SportsballGeocodeTuple(
        city="North Melbourne",
        state="VIC",
        postal="3051",
        lat=-37.7989787,
        lng=144.9388898,
        housenumber="204-206",
    ),
    "Oakland Coliseum - Oakland - CA - 94621": SportsballGeocodeTuple(
        city="Oakland",
        state="CA",
        postal="94621",
        lat=37.7514142,
        lng=-122.2034855,
        housenumber="204-206",
    ),
    "Los Angeles Memorial Coliseum - Los Angeles - CA - 90037": LOS_ANGELES_MEMORIAL_COLESIUM,
    "Western Oval - Australia": SportsballGeocodeTuple(
        city="Parkville",
        state="VIC",
        postal="3052",
        lat=-37.7781687,
        lng=144.9505349,
        housenumber="",
    ),
    "Twickenham Stadium - London": SportsballGeocodeTuple(
        city="Twickenham",
        state="",
        postal="TW2 7BA",
        lat=51.4559558,
        lng=-0.34408,
        housenumber="200",
    ),
    "Olympic Park - Australia": SportsballGeocodeTuple(
        city="Heidelberg West",
        state="VIC",
        postal="3081",
        lat=-37.7397886,
        lng=145.0317471,
        housenumber="",
    ),
    "Georgia Dome - Atlanta - GA": GEORGIA_DOME,
    "SDCCU Stadium - San Diego - CA": SDCCU_STADIUM,
    "Kardinia Park - Australia": GMHBA_STADIUM,
    "The Dome at America's Center - St. Louis - MO - 63101": TRANS_WORLD_DOME,
    "Huntington Bank Stadium - Minneapolis - MN - 55455": US_BANK_STADIUM,
    "Yarraville Oval - Australia": SportsballGeocodeTuple(
        city="Yarraville",
        state="VIC",
        postal="3013",
        lat=-37.8150875,
        lng=144.8824932,
        housenumber="",
    ),
    "Candlestick Park - San Francisco - CA": CANDLESTICK_PARK,
    "Mall of America Field - Minneapolis - MN - 55415": US_BANK_STADIUM,
    "Toorak Park - Australia": SportsballGeocodeTuple(
        city="Armadale",
        state="VIC",
        postal="3143",
        lat=-37.8528107,
        lng=145.0114987,
        housenumber="580",
    ),
    "Rogers Centre - Toronto - ON": SportsballGeocodeTuple(
        city="Toronto",
        state="ON",
        postal="M5V 1J1",
        lat=43.6415757,
        lng=-79.3942991,
        housenumber="580",
    ),
    "Euroa - Australia": SportsballGeocodeTuple(
        city="Euroa",
        state="VIC",
        postal="3666",
        lat=-36.7464868,
        lng=145.5724273,
        housenumber="",
    ),
    "Aloha Stadium - Honolulu - HI - 96818": SportsballGeocodeTuple(
        city="Aiea",
        state="HI",
        postal="96701",
        lat=21.3728016,
        lng=-157.9348645,
        housenumber="99-500",
    ),
    "Giants Stadium - East Rutherford - NJ": METLIFE_STADIUM,
    "Metrodome - Minneapolis - MN - 55415": US_BANK_STADIUM,
    "North Hobart - Australia": SportsballGeocodeTuple(
        city="North Hobart",
        state="TAS",
        postal="7000",
        lat=-42.8682555,
        lng=147.3146486,
        housenumber="1-5",
    ),
    "Texas Stadium - Irving - TX": TEXAS_STADIUM,
    "Yallourn - Australia": SportsballGeocodeTuple(
        city="Yallourn North",
        state="VIC",
        postal="3825",
        lat=-38.1629374,
        lng=146.3595115,
        housenumber="",
    ),
    "Albury - Australia": SportsballGeocodeTuple(
        city="Hamilton Valley",
        state="NSW",
        postal="2641",
        lat=-36.0349793,
        lng=146.9073391,
        housenumber="1",
    ),
    "Invesco Field - Denver - CO - 80804": EMPOWER_FIELD,
    "Brisbane Exhibition - Australia": SportsballGeocodeTuple(
        city="Bowen Hills",
        state="QLD",
        postal="4006",
        lat=-27.4503048,
        lng=153.0301872,
        housenumber="600",
    ),
    "Monster Park - San Francisco - CA - 94124": CANDLESTICK_PARK,
    "RCA Dome - Indianapolis - IN": LUCAS_OIL_STADIUM,
    "Mountain America Stadium - Tempe - AZ - 85287": MOUNTAIN_AMERICA_STADIUM,
    "Veterans Stadium - Philadelphia - PA": LINCOLN_FINANCIAL_FIELD,
    "Silverdome - Pontiac - MI": SILVERDOME,
    "Giants Stadium": METLIFE_STADIUM,
    "Mile High Stadium": EMPOWER_FIELD,
    "Pontiac Silverdome": SILVERDOME,
    "Foxboro Stadium": GILETTE_STADIUM,
    "Veterans Stadium": LINCOLN_FINANCIAL_FIELD,
    "Texas Stadium": TEXAS_STADIUM,
    "Three Rivers Stadium": ACRISURE_STADIUM,
    "Qualcomm Stadium": SDCCU_STADIUM,
    "Sun Devil Stadium": MOUNTAIN_AMERICA_STADIUM,
    "Three Rivers Stadium - Pittsburgh - PA": ACRISURE_STADIUM,
    "Moorabbin Oval - Australia": SportsballGeocodeTuple(
        city="Moorabbin",
        state="VIC",
        postal="3189",
        lat=-37.9371257,
        lng=145.0404112,
        housenumber="3",
    ),
    "Vanderbilt Stadium": SportsballGeocodeTuple(
        city="Nashville",
        state="TN",
        postal="37212",
        lat=36.1438701,
        lng=-86.8284249,
        housenumber="",
    ),
    "Liberty Bowl Memorial Stadium": LIBERTY_BOWL_MEMORIAL_STADIUM,
    "RFK Stadium": RFK_STADIUM,
    "Astrodome": NRG_STADIUM,
    "Cleveland Municipal Stadium": CLEVELAND_BROWNS_STADIUM,
    "Riverfront Stadium": PAYCOR_STADIUM,
    "Candlestick Park": CANDLESTICK_PARK,
    "Los Angeles Memorial Coliseum": LOS_ANGELES_MEMORIAL_COLESIUM,
    "Coburg Oval - Australia": SportsballGeocodeTuple(
        city="Coburg",
        state="VIC",
        postal="3058",
        lat=-37.7437207,
        lng=144.9671031,
        housenumber="",
    ),
    "Waverley Park - Australia": SportsballGeocodeTuple(
        city="Mulgrave",
        state="VIC",
        postal="3170",
        lat=-37.9258818,
        lng=145.1869239,
        housenumber="2A",
    ),
    "Rose Bowl": ROSE_BOWL,
    "Miami Orange Bowl": SportsballGeocodeTuple(
        city="Miami",
        state="FL",
        postal="33125",
        lat=25.7781487,
        lng=-80.2221747,
        housenumber="501",
    ),
    "Stanford Stadium": STANFORD_STADIUM,
    "Milwaukee County Stadium": MILWAUKEE_COUNTY_STADIUM,
    "Shea Stadium": SHEA_STADIUM,
    "Schaefer Stadium": GILETTE_STADIUM,
    "Yale Bowl": YALE_BOWL,
    "Tiger Stadium": TIGER_STADIUM,
    "War Memorial Stadium": WAR_MEMORIAL_STADIUM,
    "Yankee Stadium": YANKEE_STADIUM,
    "Municipal Stadium": MUNICIPAL_STADIUM,
    "Cotton Bowl": COTTON_BOWL,
    "Kezar Stadium": KEZAR_STADIUM,
    "Franklin Field": SportsballGeocodeTuple(
        city="Philadelphia",
        state="PA",
        postal="19104",
        lat=39.9502158,
        lng=-75.1949801,
        housenumber="235",
    ),
    "Wrigley Field": WRIGLEY_FIELD,
    "City Stadium": CITY_STADIUM,
    "IG Field - Winnipeg - MB": SportsballGeocodeTuple(
        city="Winnipeg",
        state="MB",
        postal="R3T 1Z2",
        lat=49.8077033,
        lng=-97.1456866,
        housenumber="315",
    ),
    "Subiaco - Australia": DOMAIN_STADIUM,
    "Rice Stadium": RICE_STADIUM,
    "Gabba - Australia": GABBA,
    "Independence Stadium - 71109": SportsballGeocodeTuple(
        city="Shreveport",
        state="LA",
        postal="71109",
        lat=32.475745,
        lng=-93.7944423,
        housenumber="3301",
    ),
    "Carrara - Australia": PEOPLE_FIRST_STADIUM,
    "W.A.C.A. - Australia": SportsballGeocodeTuple(
        city="East Perth",
        state="WA",
        postal="6004",
        lat=-31.959812,
        lng=115.874661,
        housenumber="",
    ),
    "Los Angeles Memorial Coliseum - Los Angeles, California - United States": LOS_ANGELES_MEMORIAL_COLESIUM,
    "Memorial Stadium (Lincoln) - Terre Haute, Indiana, USA - United States": SportsballGeocodeTuple(
        city="Lincoln",
        state="NE",
        postal="68588",
        lat=40.8206975,
        lng=-96.7081629,
        housenumber="1",
    ),
    "Kezar Stadium - San Francisco, California - United States": KEZAR_STADIUM,
    "Cotton Bowl - Dallas, Texas - United States": COTTON_BOWL,
    "Wrigley Field - Chicago, Illinois - United States": WRIGLEY_FIELD,
    "Yankee Stadium - Bronx, New York - United States": YANKEE_STADIUM,
    "Tiger Stadium - Baton Rouge, Louisiana, USA - United States": TIGER_STADIUM,
    "Lambeau Field - Green Bay, Wisconsin - United States": LAMBEAU_FIELD,
    "RFK Stadium - Washington DC - United States": RFK_STADIUM,
    "War Memorial Stadium - Little Rock, Arkansas - United States": WAR_MEMORIAL_STADIUM,
    "Three Rivers Stadium - Pittsburgh, PA - United States": ACRISURE_STADIUM,
    "Astrodome - Houston, TX - United States": NRG_STADIUM,
    "Mile High Stadium -  - ": EMPOWER_FIELD,
    "Football Park - Australia": AAMI_STADIUM,
    "Docklands - Australia": MARVEL_STADIUM,
    "Manuka Oval - Australia": MANUKA_OVAL,
    "Stadium Australia - Australia": ENGIE_STADIUM,
    "York Park - Australia": UTAS_STADIUM,
    "Marrara Oval - Australia": TIO_STADIUM,
    "Adelaide Oval - Australia": ADELAIDE_OVAL,
    "Sydney Showground - Australia": ENGIE_STADIUM,
    "Bellerive Oval - Australia": BLUNDSTONE_ARENA,
    "Cazaly's Stadium - Australia": CAZALYS_STADIUM,
    "Wellington - Australia": WESTPAC_STADIUM,
    "Traeger Park - Australia": TRAEGER_PARK,
    "Eureka Stadium - Australia": MARS_STADIUM,
    "Perth Stadium - Australia": OPTUS_STADIUM,
    "Jiangwan Stadium - Australia": JIANGWAN_STADIUM,
    "Soldier Field - Chicago, Illinois - United States": SOLDIER_FIELD,
    "Arrowhead Stadium - Kansas City, Missouri - United States": ARROWHEAD_STADIUM,
    "Yale Bowl - New Haven, Connecticut, USA - United States": YALE_BOWL,
    "Rice Stadium - Houston, Texas, U.S. - United States": RICE_STADIUM,
    "Giants Stadium - New York - United States": METLIFE_STADIUM,
    "Rose Bowl - Hampshire - England": ROSE_BOWL,
    "Riverfront Stadium - Cincinnati, OH - United States": PAYCOR_STADIUM,
    "Kingdome - Seattle, WA - United States": LUMEN_FIELD,
    "Milwaukee County Stadium - Milwaukee, WI - United States": MILWAUKEE_COUNTY_STADIUM,
    "Candlestick Park - San Francisco, CA - United States": CANDLESTICK_PARK,
    "Cleveland Municipal Stadium - Cleveland, OH - United States": CLEVELAND_BROWNS_STADIUM,
    "Stanford Stadium - Stanford, California, USA - United States": STANFORD_STADIUM,
    "Sun Devil Stadium - Tempe, Arizona, USA - United States": MOUNTAIN_AMERICA_STADIUM,
    "Liberty Bowl Memorial Stadium - Memphis, Tennessee, USA - United States": LIBERTY_BOWL_MEMORIAL_STADIUM,
    "Qualcomm Stadium - San Diego, California - United States": SDCCU_STADIUM,
    "Raymond James Stadium - Tampa Bay, Florida, USA - United States": RAYMOND_JAMES_STADIUM,
    "Cleveland Browns Stadium - Cleveland, Ohio - United States": CLEVELAND_BROWNS_STADIUM,
    "Citi Field - Flushing, New York - United States": SHEA_STADIUM,
    "Bruce Stadium - Australia": SportsballGeocodeTuple(
        city="Bruce",
        state="ACT",
        postal="2617",
        lat=-35.249994,
        lng=149.1001351,
        housenumber="",
    ),
    "Blacktown - Australia": BLACKTOWN_PARK,
    "Gillette Stadium - Foxborough, Massachusetts - United States": GILETTE_STADIUM,
    "Riverway Stadium - Australia": RIVERWAY_STADIUM,
    "Norwood Oval - Australia": NORWOOD_OVAL,
    "Summit Sports Park - Australia": ADELAIDE_HILLS,
    "Mile High": EMPOWER_FIELD,
    "Nissan Stadium - Nashville, Tennessee, USA - United States": NISSAN_STADIUM,
    "SoFi Stadium - Inglewood, CA - United States": SOFI_STADIUM,
    "Highmark Stadium - Orchard Park, New York - United States": HIGHMARK_STADIUM,
    "Mercedes-Benz Stadium - Atlanta, Georgia - United States": MERCEDES_BANZ_STADIUM,
    "Lucas Oil Stadium - Indianapolis, Indiana, USA - United States": LUCAS_OIL_STADIUM,
    "U.S. Bank Stadium - Minneapolis, Minnesota, USA - United States": US_BANK_STADIUM,
    "Caesars Superdome - New Orleans, Louisiana, United States - United States": CAESARS_SUPERDOME,
    "Acrisure Stadium - Pittsburgh, Pennsylvania, USA - United States": ACRISURE_STADIUM,
    "FedExField - Landover, Maryland, USA - United States": FEDEX_FIELD,
    "M&T Bank Stadium - Baltimore, Maryland - United States": MTT_BANK_STADIUM,
    "Empower Field at Mile High - Denver, Colorado, USA - United States": EMPOWER_FIELD,
    "Lumen Field - Seattle, WA, USA - United States": LUMEN_FIELD,
    "MetLife Stadium - East Rutherford, New Jersey - United States": METLIFE_STADIUM,
    "Lincoln Financial Field - Philadelphia, Pennsylvania - United States": LINCOLN_FINANCIAL_FIELD,
    "Paycor Stadium - Cincinnati, Ohio - United States": PAYCOR_STADIUM,
    "Ford Field - Detroit, Michigan - United States": FORD_FIELD,
    "EverBank Stadium - Jacksonville, Florida - United States": EVERBANK_STADIUM,
    "NRG Stadium - Houston, Texas, USA - United States": NRG_STADIUM,
    "State Farm Stadium - Glendale, Arizona - United States": STATE_FARM_STADIUM,
    "AT&T Stadium - Arlington, Texas, USA - United States": ATT_STADIUM,
    "Bank of America Stadium - Charlotte, NC, United States - United States": BANK_OF_AMERICA_STADIUM,
    "Hard Rock Stadium - Coral Gables, Florida, USA - United States": HARD_ROCK_STADIUM,
    "GEHA Field at Arrowhead Stadium - Kansas City, Missouri, United States - United States": ARROWHEAD_STADIUM,
    "Wembley Stadium - Wembley, London - International": WEMBLEY_STADIUM,
    "Tottenham Hotspur Stadium - Bill Nicholson Way, Tottenham, London - England": TOTTENHAM_HOTSPURS_STADIUM,
    "Deutsche Bank Park - Frankfurt, Germany - Germany": DEUTSCHE_BANK_PARK,
    "Arlington Stadium - Arlington, TX - United States": ATT_STADIUM,
    "Allegiant Stadium - Paradise, Nevada, USA - United States": ALLEGIANT_STADIUM,
    "Neo Química Arena - São Paulo, Brazil - Brazil": NEO_QUIMICA_ARENA,
    "Allianz Arena - Munich, Germany - Germany": ALLIANZ_ARENA,
    "Levi's Stadium - Santa Clara, California, USA - United States": LEVI_STADIUM,
    "MCG": MCG,
    "SCG": SCG,
    "ENGIE Stadium": ENGIE_STADIUM,
    "Adelaide Oval": ADELAIDE_OVAL,
    "Gabba": GABBA,
    "Optus Stadium": OPTUS_STADIUM,
    "Marvel Stadium": MARVEL_STADIUM,
    "Mars Stadium": MARS_STADIUM,
    "GMHBA Stadium": GMHBA_STADIUM,
    "UTAS Stadium": UTAS_STADIUM,
    "People First Stadium": PEOPLE_FIRST_STADIUM,
    "Blundstone Arena": BLUNDSTONE_ARENA,
    "Manuka Oval": MANUKA_OVAL,
    "TIO Stadium": TIO_STADIUM,
    "Norwood Oval": NORWOOD_OVAL,
    "Adelaide Hills": ADELAIDE_HILLS,
    "Traeger Park": TRAEGER_PARK,
    "Cazaly’s Stadium": CAZALYS_STADIUM,
    "Marvl": MARVEL_STADIUM,
    "Accor Stadium": ENGIE_STADIUM,
    "Riverway Stadium": RIVERWAY_STADIUM,
    "Jiangwan Sports Centre": JIANGWAN_STADIUM,
    "Domain Stadium": DOMAIN_STADIUM,
    "Westpac Stadium": WESTPAC_STADIUM,
    "AAMI Stadium": AAMI_STADIUM,
    "Blacktown Park": BLACKTOWN_PARK,
}


@MEMORY.cache(ignore=["session"])
def create_google_address_model(
    query: str, session: requests.Session, dt: datetime.datetime
) -> AddressModel:
    """Create address model from google."""
    g = _CACHED_GEOCODES.get(query)
    if g is None:
        logging.warning("Failed to find query: %s", query)
        g = geocoder.google(query, session=session)
        _CACHED_GEOCODES[query] = g
    latitude = g.lat
    longitude = g.lng
    weather_model = None
    tz = "UTC"
    if latitude is not None and longitude is not None:
        tf = TimezoneFinder()
        timezone = tf.timezone_at(lng=longitude, lat=latitude)
        if timezone is not None:
            tz = timezone
        weather_model = create_openmeteo_weather_model(
            session,
            latitude,
            longitude,
            dt,  # pyright: ignore
            tz,
        )
    return AddressModel(
        city=g.city,
        state=g.state,
        zipcode=g.postal,
        latitude=latitude,
        longitude=longitude,
        housenumber=g.housenumber,
        weather=weather_model,  # pyright: ignore
        timezone=tz,
    )



"""A weight based on no bias towards the current time."""

import numpy as np
import pandas as pd

from .weight import Weight


class NoopWeight(Weight):
    """Noop weight class."""

    @classmethod
    def name(cls) -> str:
        """The name of the weight class."""
        return "noop"

    def process(self, y: pd.DataFrame) -> np.ndarray:
        return np.array([1.0 for _ in range(len(y))])



"""Tests for the catboost trainer."""
import unittest

from sportsball.strategy.trainers.catboost import CatboostTrainer


class TestCatboostTrainer(unittest.TestCase):

    def setUp(self):
        self._trainer = CatboostTrainer("", [], [])

    def test_not_null(self):
        self.assertIsNotNone(self._trainer)



"""Tests for the total feature class."""
import unittest

import pandas as pd
from sportsball.strategy.features.columns import \
    team_identifier_column, player_identifier_column, player_column_prefix, venue_identifier_column, team_column_prefix, kick_column
from sportsball.strategy.features.total_feature import TotalFeature, TOTAL_COLUMN_PREFIX, TOTAL_GAMES_COLUMN, TOTAL_KICKS_COLUMN
from sportsball.data.game_model import VENUE_COLUMN_PREFIX
from sportsball.data.league_model import DELIMITER


class TestTotalFeature(unittest.TestCase):

    def setUp(self):
        self._total_feature = TotalFeature()

    def test_team_venue_total_games(self):
        df = pd.DataFrame(data={
            team_identifier_column(0): ["t1", "t1", "t1"],
            team_identifier_column(1): ["t2", "t2", "t2"],
            venue_identifier_column(): ["v1", "v1", "v1"],
            kick_column(0, 0): [0, 1, 2],
            kick_column(0, 1): [1, 2, 3],
            kick_column(1, 0): [3, 4, 5],
            kick_column(1, 1): [5, 6, 7],
        })
        df = self._total_feature.process(df)
        total_games = df[DELIMITER.join([TOTAL_COLUMN_PREFIX, team_column_prefix(0), VENUE_COLUMN_PREFIX, TOTAL_GAMES_COLUMN])].values.tolist()
        self.assertListEqual(total_games, [0, 1, 2])



"""ESPN league model."""

from typing import Any, Iterator

import requests

from ..game_model import GameModel
from ..league import League
from ..league_model import LeagueModel
from ..season_type import SeasonType
from .espn_game_model import create_espn_game_model


def _season_type_from_name(name: str) -> SeasonType:
    if name == "Regular Season":
        return SeasonType.REGULAR
    if name == "Preseason":
        return SeasonType.PRESEASON
    if name == "Postseason":
        return SeasonType.POSTSEASON
    if name == "Off Season":
        return SeasonType.OFFSEASON
    raise ValueError(f"Unrecognised season name: {name}")


class ESPNLeagueModel(LeagueModel):
    """ESPN implementation of the league model."""

    def __init__(
        self, start_url: str, league: League, session: requests.Session
    ) -> None:
        super().__init__(league, session)
        self._start_url = start_url

    def _produce_games(
        self, week: dict[str, Any], week_count: int, season_type_json: dict[str, Any]
    ) -> Iterator[GameModel]:
        events_page = 1
        events_count = 0
        while True:
            if "events" not in week:
                break
            events_response = self.session.get(
                week["events"]["$ref"] + f"&page={events_page}"
            )
            events = events_response.json()
            for event_item in events["items"]:
                event_response = self.session.get(event_item["$ref"])
                event_response.raise_for_status()
                event = event_response.json()
                yield create_espn_game_model(
                    event,
                    week_count,
                    events_count,
                    self.session,
                    self.league,
                    season_type_json["year"],
                    _season_type_from_name(season_type_json["name"]),
                )
                events_count += 1
            if events_page >= events["pageCount"]:
                break
            events_page += 1

    def _produce_week_games(
        self, season_type_json: dict[str, Any], page: int
    ) -> Iterator[GameModel]:
        game_page = 1
        week_count = 0
        while True:
            weeks_response = self.session.get(
                season_type_json["weeks"]["$ref"] + f"&page={page}"
            )
            weeks_response.raise_for_status()
            weeks = weeks_response.json()
            for item in weeks["items"]:
                week_response = self.session.get(item["$ref"])
                week_response.raise_for_status()
                week = week_response.json()
                yield from self._produce_games(week, week_count, season_type_json)
                week_count += 1
            if game_page >= weeks["pageCount"]:
                break
            game_page += 1

    @property
    def games(self) -> Iterator[GameModel]:
        page = 1
        while True:
            response = self.session.get(self._start_url + f"&page={page}")
            response.raise_for_status()
            seasons = response.json()
            for item in seasons["items"]:
                season_response = self.session.get(item["$ref"])
                season_response.raise_for_status()
                season_json = season_response.json()

                for season_item in season_json["types"]["items"]:
                    season_type_response = self.session.get(season_item["$ref"])
                    season_type_response.raise_for_status()
                    season_type_json = season_type_response.json()

                    yield from self._produce_week_games(season_type_json, page)

            if page >= seasons["pageCount"]:
                break
            page += 1



"""AFL aussportsbetting odds model."""

from ....cache import MEMORY
from ...team_model import OddsModel
from .afl_aussportsbetting_bookie_model import \
    create_afl_aussportsbetting_bookie_model


@MEMORY.cache
def create_afl_aussportsbetting_odds_model(odds: float) -> OddsModel:
    """Create an odds model based off aus sports betting."""
    bookie = create_afl_aussportsbetting_bookie_model()
    return OddsModel(odds=odds, bookie=bookie)  # pyright: ignore



"""NFL ESPN league model."""

import requests

from ...espn.espn_league_model import ESPNLeagueModel
from ...league import League

_SEASON_URL = (
    "https://sports.core.api.espn.com/v2/sports/football/leagues/nfl/seasons?limit=100"
)


class NFLESPNLeagueModel(ESPNLeagueModel):
    """NFL ESPN implementation of the league model."""

    def __init__(self, session: requests.Session) -> None:
        super().__init__(_SEASON_URL, League.NFL, session)



"""AFL AFLTables league model."""

import os
import urllib.parse
from typing import Iterator
from urllib.parse import urlparse

import requests
import tqdm
from bs4 import BeautifulSoup

from ...game_model import GameModel
from ...league import League
from ...league_model import LeagueModel
from ...season_type import SeasonType
from .afl_afltables_game_model import create_afl_afltables_game_model

_SEASON_URL = "https://afltables.com/afl/seas/season_idx.html"


class AFLAFLTablesLeagueModel(LeagueModel):
    """AFL AFLTables implementation of the league model."""

    def __init__(self, session: requests.Session) -> None:
        super().__init__(League.AFL, session)

    def _produce_games(
        self, season_url: str, season_type: SeasonType
    ) -> Iterator[GameModel]:
        # pylint: disable=too-many-branches,too-many-locals
        o = urlparse(season_url)
        last_component = o.path.split("/")[-1]
        filename, _ = os.path.splitext(last_component)
        year = int(filename)

        response = self._session.get(season_url)
        soup = BeautifulSoup(response.text, "html.parser")
        in_finals = False
        game_number = 0
        last_round_number = 0
        urls_duplicates = set()
        last_ladder_ranks: dict[str, int] = {}
        for table in soup.find_all("table"):
            for b in table.find_all("b"):
                if b.get_text() == "Finals":
                    in_finals = True
                    break
            for a in table.find_all("a", href=True):
                if a.get_text().strip().lower() == "match stats":
                    url = urllib.parse.urljoin(season_url, a.get("href"))
                    if url in urls_duplicates:
                        continue
                    if not in_finals and season_type == SeasonType.REGULAR:
                        model = create_afl_afltables_game_model(
                            game_number,
                            self._session,
                            url,
                            last_round_number,
                            last_ladder_ranks,
                            self.league,  # pyright: ignore
                            year,
                            season_type,
                        )
                        model_week = model.week
                        if model_week is None:
                            raise ValueError("model_week is null")
                        last_round_number = model_week  # pyright: ignore
                        if season_type == SeasonType.REGULAR:
                            yield model  # pyright: ignore
                    elif in_finals and season_type == SeasonType.POSTSEASON:
                        yield create_afl_afltables_game_model(
                            game_number,
                            self._session,
                            url,
                            last_round_number,
                            None,
                            self.league,  # pyright: ignore
                            year,
                            season_type,
                        )
                    game_number += 1
                    urls_duplicates.add(url)
            ladder_count = None
            for tr in table.find_all("tr"):
                if ladder_count is None:
                    for td in tr.find_all("td"):
                        if (
                            td.get_text().strip().lower()
                            == f"rd {last_round_number} ladder"
                        ):
                            ladder_count = 0
                            last_ladder_ranks = {}
                            break
                else:
                    for td in tr.find_all("td"):
                        team_short_name = td.get_text().strip().upper()
                        if len(team_short_name) > 2 and team_short_name.isalpha():
                            continue
                        ladder_count += 1
                        last_ladder_ranks[team_short_name] = ladder_count
                        break

    @property
    def games(self) -> Iterator[GameModel]:
        response = self.session.get(_SEASON_URL)
        soup = BeautifulSoup(response.text, "html.parser")
        with tqdm.tqdm(desc="AFLTables seasons") as pbar:
            for table in soup.find_all("table"):
                for tr in table.find_all("tr"):
                    for td in tr.find_all("td"):
                        for a in td.find_all("a"):
                            url = urllib.parse.urljoin(_SEASON_URL, a.get("href"))
                            yield from self._produce_games(url, SeasonType.REGULAR)
                            yield from self._produce_games(url, SeasonType.POSTSEASON)
                            pbar.update(2)



"""Combined venue model."""

from ..venue_model import VenueModel
from .combined_address_model import create_combined_address_model


def create_combined_venue_model(
    venue_models: list[VenueModel], identifier: str | None
) -> VenueModel | None:
    """Create a venue model by combining many venue models."""
    if not venue_models or identifier is None:
        return None
    address_models = []
    is_grass = None
    is_indoor = None
    for venue_model in venue_models:
        venue_model_address = venue_model.address
        if venue_model_address is not None:
            address_models.append(venue_model_address)
        venue_model_is_grass = venue_model.is_grass
        if venue_model_is_grass is not None:
            is_grass = venue_model_is_grass
        venue_model_is_indoor = venue_model.is_indoor
        if venue_model_is_indoor is not None:
            is_indoor = venue_model_is_indoor
    return VenueModel(
        identifier=identifier,
        name=venue_models[0].name,
        address=create_combined_address_model(address_models),  # pyright: ignore
        is_grass=is_grass,
        is_indoor=is_indoor,
    )



"""The main sportsball class for accessing data."""

# pylint: disable=too-few-public-methods
import datetime
from typing import Dict
from warnings import simplefilter

import pandas as pd
import requests
import requests_cache
from dotenv import load_dotenv
from retry_requests import retry  # type: ignore

from .data.afl import AFLLeagueModel
from .data.league import League
from .data.league_model import LeagueModel
from .data.nba import NBALeagueModel
from .data.ncaaf import NCAAFLeagueModel
from .data.nfl import NFLLeagueModel
from .portfolio import Portfolio
from .strategy import Strategy


class SportsBall:
    """The main sportsball class."""

    _leagues: Dict[str, LeagueModel]
    _session: requests.Session

    def __init__(self) -> None:
        cache_session = requests_cache.CachedSession(
            "sportsball",
            expire_after=datetime.timedelta(days=365),
        )
        self._session = retry(cache_session, retries=5, backoff_factor=0.2)
        self._leagues = {}
        simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
        load_dotenv()

    def league(self, league: League) -> LeagueModel:
        """Provide a league model for the given league."""
        if league not in self._leagues:
            if league == League.NFL:
                self._leagues[league] = NFLLeagueModel(self._session)
            elif league == League.AFL:
                self._leagues[league] = AFLLeagueModel(self._session)
            elif league == League.NBA:
                self._leagues[league] = NBALeagueModel(self._session)
            elif league == League.NCAAF:
                self._leagues[league] = NCAAFLeagueModel(self._session)
            else:
                raise ValueError(f"Unrecognised league: {league}")
        return self._leagues[league]

    def create_strategy(self, league: LeagueModel, name: str) -> Strategy:
        """Creates a strategy."""
        return Strategy(league.to_frame(), name)

    def create_portfolio(self, strategies: list[Strategy], name: str) -> Portfolio:
        """Creates a portfolio."""
        return Portfolio(strategies, name)



"""A class for holding weather information."""

from pydantic import BaseModel


class WeatherModel(BaseModel):
    """The serialisable class for representing weather."""

    temperature: float | None



"""The prototype class for a reducer."""

import pandas as pd


class Reducer:
    """The prototype reducer class."""

    # pylint: disable=too-few-public-methods

    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """Process the dataframe and remove the necessary features."""
        raise NotImplementedError("process is not implemented in parent class.")



"""ESPN odds model."""

from typing import Any

from ...cache import MEMORY
from ..bookie_model import BookieModel
from ..team_model import OddsModel

MONEYLINE_KEY = "moneyLine"


@MEMORY.cache
def create_espn_odds_model(odds: dict[str, Any], bookie: BookieModel) -> OddsModel:
    """Create an odds model with ESPN."""
    odds_val = 0.0
    moneyline = odds[MONEYLINE_KEY]
    if moneyline > 0:
        odds_val = (float(moneyline) / 100.0) + 1.0
    elif moneyline < 0:
        odds_val = (100.0 / float(abs(moneyline))) + 1.0
    return OddsModel(odds=odds_val, bookie=bookie)



"""NFL SportsDB game model."""

# pylint: disable=too-many-arguments
import datetime
from typing import Any

import requests
from dateutil import parser

from ....cache import MEMORY
from ...game_model import GameModel
from ...league import League
from ...season_type import SeasonType
from .nfl_sportsdb_team_model import create_nfl_sportsdb_team_model
from .nfl_sportsdb_venue_model import create_nfl_sportsdb_venue_model


@MEMORY.cache(ignore=["session"])
def create_nfl_sportsdb_game_model(
    session: requests.Session,
    game: dict[str, Any],
    week_number: int,
    game_number: int,
    league: League,
    year: int | None,
    season_type: SeasonType | None,
) -> GameModel:
    """Create an NFL SportsDB game model."""
    try:
        dt = datetime.datetime.fromisoformat(game["strTimestamp"])
    except TypeError:
        dt = parser.parse(game["dateEvent"])
    venue = create_nfl_sportsdb_venue_model(session, game["idVenue"], dt)
    home_score = float(game["intHomeScore"] if game["intHomeScore"] is not None else 0)
    away_score = float(game["intAwayScore"] if game["intAwayScore"] is not None else 0)
    teams = [
        create_nfl_sportsdb_team_model(
            game["idHomeTeam"], game["strHomeTeam"], home_score
        ),
        create_nfl_sportsdb_team_model(
            game["idAwayTeam"], game["strAwayTeam"], away_score
        ),
    ]
    return GameModel(
        dt=dt,
        week=week_number,
        game_number=game_number,
        venue=venue,  # pyright: ignore
        teams=teams,  # pyright: ignore
        league=league,
        year=year,
        season_type=season_type,
        end_dt=None,
        attendance=None,
    )
